# Copyright 2012 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from chirp.drivers import icf
from chirp import chirp_common, directory
from chirp import bitwise

mem_format = """
struct memory {
  bbcd freq[3];
  bbcd offset[3];
  u8 rtone;
  u8 ctone;
};

struct flags {
  u8 empty:1,
     skip:1,
     tmode:2,
     duplex:2,
     unknown2:2;
};

struct memory memory[100];

#seekto 0x0400;
struct {
  char name[4];
} names[100];

#seekto 0x0600;
struct flags flags[100];
"""

DUPLEX = ["", "", "-", "+"]
TMODES = ["", "", "Tone", "TSQL"]


def _get_freq(bcd_array):
    lastnibble = bcd_array[2].get_bits(0x0F)
    return (int(bcd_array) - lastnibble) * 1000 + lastnibble * 500


def _set_freq(bcd_array, freq):
    bitwise.int_to_bcd(bcd_array, freq / 1000)
    bcd_array[2].set_raw(bcd_array[2].get_bits(0xF0) + freq % 10000 // 500)


@directory.register
class ICT8ARadio(icf.IcomCloneModeRadio):
    """Icom IC-T8A"""
    VENDOR = "Icom"
    MODEL = "IC-T8A"

    _model = "\x19\x03\x00\x01"
    _memsize = 0x07B0
    _endframe = "Icom Inc\x2e"

    _ranges = [(0x0000, 0x07B0, 16)]

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.valid_tmodes = TMODES
        rf.valid_duplexes = DUPLEX
        rf.valid_bands = [(50000000, 54000000),
                          (118000000, 174000000),
                          (400000000, 470000000)]
        rf.valid_skips = ["", "S"]
        rf.valid_modes = ["FM"]
        rf.memory_bounds = (0, 99)
        rf.valid_name_length = 4
        rf.valid_characters = chirp_common.CHARSET_UPPER_NUMERIC
        rf.has_dtcs = False
        rf.has_dtcs_polarity = False
        rf.valid_tuning_steps = [5, 10, 12.5, 15, 20, 25, 30, 50, 100]
        rf.has_tuning_step = False
        rf.has_mode = False
        rf.has_bank = False
        return rf

    def process_mmap(self):
        self._memobj = bitwise.parse(mem_format, self._mmap)

    def get_raw_memory(self, number):
        return (str(self._memobj.memory[number]) +
                str(self._memobj.names[number]) +
                str(self._memobj.flags[number]))

    def get_memory(self, number):
        _mem = self._memobj.memory[number]
        _flg = self._memobj.flags[number]
        _name = self._memobj.names[number]

        mem = chirp_common.Memory()
        mem.number = number

        if _flg.empty:
            mem.empty = True
            return mem

        mem.freq = _get_freq(_mem.freq)
        mem.offset = _get_freq(_mem.offset)
        mem.rtone = chirp_common.TONES[_mem.rtone - 1]
        mem.ctone = chirp_common.TONES[_mem.ctone - 1]
        mem.duplex = DUPLEX[_flg.duplex]
        mem.tmode = TMODES[_flg.tmode]
        mem.skip = _flg.skip and "S" or ""
        if _name.name.get_raw(asbytes=False) != "\xFF\xFF\xFF\xFF":
            mem.name = str(_name.name).rstrip()

        return mem

    def set_memory(self, mem):
        _mem = self._memobj.memory[mem.number]
        _flg = self._memobj.flags[mem.number]
        _name = self._memobj.names[mem.number]

        if mem.empty:
            _flg.empty = True
            return

        _mem.set_raw("\x00" * 8)
        _flg.set_raw("\x00")

        _set_freq(_mem.freq, mem.freq)
        _set_freq(_mem.offset, mem.offset)
        _mem.rtone = chirp_common.TONES.index(mem.rtone) + 1
        _mem.ctone = chirp_common.TONES.index(mem.ctone) + 1
        _flg.duplex = DUPLEX.index(mem.duplex)
        _flg.tmode = TMODES.index(mem.tmode)
        _flg.skip = mem.skip == "S"

        if mem.name:
            _name.name = mem.name.ljust(4)
        else:
            _name.name = "\xFF\xFF\xFF\xFF"



# Copyright 2010 Dan Smith <dsmith@danplanet.com>
# Copyright 2017 Nicolas Pike <nick@zbm2.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from chirp.drivers import yaesu_clone
from chirp import chirp_common, directory, bitwise
from chirp import errors
from chirp import memmap
from chirp.settings import RadioSettingGroup, RadioSetting, RadioSettings, \
    RadioSettingValueString, RadioSettingValueList, \
    InvalidValueError
from chirp import util

import string
LOG = logging.getLogger(__name__)

# Testing

# 37    PAG.ABK     Turn the pager answer back Function ON/OFF
# 38    PAG.CDR     Specify a personal code (receive)
# 39    PAG.CDT     Specify a personal code (transmit)
# 47    RX.MOD      Select the receive mode. Auto FM AM

MEM_SETTINGS_FORMAT = """

// FT-70DE New Model #5329
//
// Communications Mode ? AMS,FM DN,DW   TX vs RX?
// Mode not currently correctly stored in memories ? - ALL show as FM in memories
// SKIP test/where stored
// Check storage of steps
// Pager settings ?
// Triple check/ understand _memsize and _block_lengths
// Bank name label name size display 6 store 16? padded with 0xFF same for MYCALL and message
// CHIRP mode DIG not supported - is there a CHIRP Fusion mode? Auto?
// Check character set
// Supported Modes ?
// Supported Bands ?
// rf.has_dtcs_polarity = False - think radio supports DTCS polarity
// rf.memory_bounds = (1, 900) - should this be 0? as zero displays as blank
// RPT offsets (stored per band) not included.
// 59 Display radio firmware version info and Radio ID?
// Front Panel settings power etc?
// Banks and VFO?

// Features Required
// Default AMS and Memory name (in mem extras) to enabled.

// Bugs
// MYCALL and Opening Message errors if not 10 characters
// Values greater than one sometimes stored as whole bytes, these need to be refactored into bit fields
// to prevent accidental overwriting of adjacent values
// Bank Name length not checked on gui input - but first 6 characters are saved correctly.
// Extended characters entered as bank names on radio are corrupted in Chirp

// Missing
// 50 SCV.WTH Set the VFO scan frequency range. BAND / ALL - NOT FOUND
// 49 SCM.WTH Set the memory scan frequency range. ALL / BAND - NOT FOUND

// Radio Questions
// Temp unit C/F not saved by radio, always goes back to C ?
// 44 RF SQL Adjusts the RF Squelch threshold level. OFF / S1 - S9? Default is OFF - Based on RF strength - for AM? How
// is this different from F, Monitor, Dial Squelch?
// Password setting on radio allows letters (DTMF), but letters cannot be entered at the radio's password prompt?
// 49 SCM.WTH Set the memory scan frequency range. ALL / BAND Defaults to ALL Not Band as stated in the manual.

   #seekto 0x049a;
    struct {
    u8 unknown0:4,
    squelch:4;              // Squelch F, Monitor, Dial Adjust the Squelch level
    } squelch_settings;

    #seekto 0x04ba;
    struct {
    u8 unknown:3,
    scan_resume:5;          // 52 SCN.RSM   Configure the scan stop mode settings. 2.0 S - 5.0 S - 10.0 S / BUSY / HOLD
    u8 unknown1:3,
    dw_resume_interval:5;   // 22 DW RSM    Configure the scan stop mode settings for Dual Receive. 2.0S-10.0S/BUSY/HOLD
    u8 unknown2;
    u8 unknown3:3,
    apo:5;                  // 02 APO       Set the length of time until the transceiver turns off automatically.
    u8 unknown4:6,
    gm_ring:2;              // 24 GM RNG    Select the beep option while receiving digital GM info. OFF/IN RNG/ALWAYS
    u8 temp_cf;             // Placeholder as not found
    u8 unknown5;
    } first_settings;

    #seekto 0x04ed;
    struct {
    u8 unknown1:1,
    unknown2:1,
    unknown3:1,
    unknown4:1,
    unknown5:1,
    unknown6:1,
    unknown7:1,
    unknown8:1;
    } test_bit_field;

    #seekto 0x04c0;
    struct {
    u8 unknown1:5,
    beep_level:3;           // 05 BEP.LVL   Beep volume setting LEVEL1 - LEVEL4 - LEVEL7
    u8 unknown2:6,
    beep_select:2;          // 04 BEEP      Sets the beep sound function OFF / KEY+SC / KEY
    } beep_settings;

    #seekto 0x04ce;
    struct {
    u8 lcd_dimmer;                      // 14 DIMMER    LCD Dimmer
    u8 dtmf_delay;                      // 18 DT DLY    DTMF delay
    u8 unknown0[3];
    u8 unknown1:4,
    unknown1:4;
    u8 lamp;                            // 28 LAMP      Set the duration time of the backlight and keys to be lit
    u8 lock;                            // 30 LOCK      Configure the lock mode setting. KEY/DIAL/K+D/PTT/K+P/D+P/ALL
    u8 unknown2_1;
    u8 mic_gain;                        // 31 MCGAIN    Adjust the microphone gain level
    u8 unknown2_3;
    u8 dw_interval;                     // 21 DW INT Set the priority memory ch mon int during Dual RX 0.1S-5.0S-10.0S
    u8 ptt_delay;                       // 42 PTT.DLY   Set the PTT delay time. OFF / 20 ms / 50 ms / 100 ms / 200 ms
    u8 rx_save;                         // 48 RX.SAVE   Set the battery save time. OFF / 0.2 s - 60.0 s
    u8 scan_restart;                    // 53 SCN.STR   Set the scanning restart time.  0.1 s - 2.0 s - 10.0 s
    u8 unknown2_5;
    u8 unknown2_6;
    u8 unknown4[5];
    u8 tot;                             // 56 TOT       Set the transmission timeout timer
    u8 unknown5[3];          // 26
    u8 vfo_mode:1,                      // 60 VFO.MOD   Set freq setting range in the VFO mode by DIAL knob. ALL / BAND
    unknown7:1,
    scan_lamp:1,                        // 51 SCN.LMP   Set the scan lamp ON or OFF when scanning stops On/Off
    unknown8:1,
    ars:1,                              // 45 RPT.ARS   Turn the ARS function on/off.
    dtmf_speed:1,                       // 20 DT SPD    Set DTMF speed
    unknown8:1,
    dtmf_mode:1;                        // DTMF Mode set from front panel
    u8 busy_led:1,                      // Not Supported ?
    unknown8_2:1,
    unknown8_3:1,
    bclo:1,                             // 03 BCLO      Turns the busy channel lockout function on/off.
    beep_edge:1,                        // 06 BEP.Edg   Sets the beep sound ON or OFF when a band edge is encountered.
    unknown8_6:1,
    unknown8_7:1,
    unknown8_8:1;            // 28
    u8 unknown9_1:1,
    unknown9_2:1,
    unknown9_3:1,
    unknown9_4:1,
    unknown9_5:1,
    password:1,                         // Placeholder location
    home_rev:1,                         // 26 HOME/REV   Select the function of the [HOME/REV] key.
    moni:1;                             // 32 Mon/T-Call Select the function of the [MONI/T-CALL] switch.
    u8 gm_interval:4,       // 30       // 25 GM INT Set tx interval of digital GM information. OFF / NORMAL / LONG
    unknown10:4;
    u8 unknown11;
    u8 unknown12:1,
    unknown12_2:1,
    unknown12_3:1,
    unknown12_4:1,
    home_vfo:1,                         // 27 HOME->VFO  Turn transfer VFO to the Home channel ON or OFF.
    unknown12_6:1,
    unknown12_7:1,
    dw_rt:1;                // 32       // 23 DW RVT Turn "Priority Channel Revert" feature ON or OFF during Dual Rx.
    u8 unknown33;
    u8 unknown34;
    u8 unknown35;
    u8 unknown36;
    u8 unknown37;
    u8 unknown38;
    u8 unknown39;
    u8 unknown40;
    u8 unknown41;
    u8 unknown42;
    u8 unknown43;
    u8 unknown44;
    u8 unknown45;
    u8 prog_key1;           // P1 Set Mode Items to the Programmable Key
    u8 prog_key2;           // P2 Set Mode Items to the Programmable Key
    u8 unknown48;
    u8 unknown49;
    u8 unknown50;
    } scan_settings;

    #seekto 0x064b;
    struct {
    u8 unknown1:1,
    unknown2:1,
    unknown3:1,
    unknown4:1,
    vfo_scan_width:1,       // Placeholder as not found - 50 SCV.WTH Set the VFO scan frequency range. BAND / ALL
    memory_scan_width:1,    // Placeholder as not found - 49 SCM.WTH Set the memory scan frequency range. ALL / BAND
    unknown7:1,
    unknown8:1;
    } scan_settings_1;

    #seekto 0x06B6;
    struct {
    u8 unknown1:3,
    volume:5;               // # VOL and Dial  Adjust the volume level
    } scan_settings_2;

    #seekto 0x0690;         // Memory or VFO Settings Map?
    struct {
    u8 unknown[48];         // Array cannot be 64 elements!
    u8 unknown1[16];        // Exception: Not implemented for chirp.bitwise.structDataElement
    } vfo_info_1;

    #seekto 0x0710;         // Backup Memory or VFO Settings Map?
    struct {
    u8 unknown[48];
    u8 unknown1[16];
    } vfo_backup_info_1;

    #seekto 0x047e;
    struct {
    u8 unknown1;
    u8 flag;
    u16 unknown2;
    struct {
    char padded_string[6];              // 36 OPN.MSG   Select MSG then key vm to edit it
    } message;
    } opening_message;                  // 36 OPN.MSG   Select the Opening Message when transceiver is ON. OFF/MSG/DC

    #seekto 0x094a;                     // DTMF Memories
    struct {
    u8 memory[16];
    } dtmf[10];

    #seekto 0x154a;
    struct {
    u16 channel[100];
    } bank_members[24];

    #seekto 0x54a;
    struct {
    u16 in_use;
    } bank_used[24];

    #seekto 0x0EFE;
    struct {
    u8 unknown[2];
    u8 name[6];
    u8 unknown1[10];
    } bank_info[24];

    #seekto 0xCF30;
    struct {
    u8 unknown0;
    u8 unknown1;
    u8 unknown2;
    u8 unknown3;
    u8 unknown4;
    u8 unknown5;
    u8 unknown6;
    u8 digital_popup;                   // 15 DIG.POP   Call sign display pop up time
    } digital_settings_more;

    #seekto 0xCF7C;
    struct {
    u8 unknown0:6,
    ams_tx_mode:2;                      // AMS TX Mode  Short Press AMS button AMS TX Mode
    u8 unknown1;
    u8 unknown2:7,
    standby_beep:1;                     // 07 BEP.STB   Standby Beep in the digital C4FM mode. On/Off
    u8 unknown3;
    u8 unknown4:6,
    gm_ring:2;                          // 24 GM RNG Select beep option while rx digital GM info. OFF/IN RNG/ALWAYS
    u8 unknown5;
    u8 rx_dg_id;                        // RX DG-ID     Long Press Mode Key, Mode Key to select, Dial
    u8 tx_dg_id;                        // TX DG-ID     Long Press Mode Key, Dial
    u8 unknown6:7,
    vw_mode:1;                          // 16 DIG VW    Turn the VW mode selection ON or OFF
    u8 unknown7;
    } digital_settings;

    // ^^^ All above referenced U8's have been refactored to minimum number of bits.

    """

MEM_FORMAT = """
    #seekto 0x2D4A;
    struct {                            // 32 Bytes per memory entry
    u8 display_tag:1,                   // 0 Display Freq, 1 Display Name
    unknown0:1,                         // Mode if AMS not selected????????
    deviation:1,                        // 0 Full deviation (Wide), 1 Half deviation (Narrow)
    clock_shift:1,                      // 0 None, 1 CPU clock shifted
    unknown1:4;                                                                         // 1
    u8 mode:2,                          // FM,AM,WFM only? - check
    duplex:2,                           // Works
    tune_step:4;                        // Works - check all steps? 7 = Auto            // 1
    bbcd freq[3];                       // Works                                        // 3
    u8 power:2,                         // Works
    unknown2:1,                         // 0 FM, 1 Digital - If AMS off
    ams:1,                              // 0 AMS off, 1 AMS on ?
    tone_mode:4;                        // Works                                        // 1
    u8 charsetbits[2];                                                                  // 2
    char label[6];                      // Works - Can only input 6 on screen           // 6
    char unknown7[10];                  // Rest of label ???                            // 10
    bbcd offset[3];                     // Works                                        // 3
    u8 unknown5:2,
    tone:6;                             // Works                                       // 1
    u8 unknown6:1,
    dcs:7;                              // Works                                        // 1
    u8 unknown9;
    u8 ams_on_dn_vw_fm:2,               // AMS DN, AMS VW, AMS FM
    unknown8_3:1,
    unknown8_4:1,
    smeter:4;
    u8 unknown10:2,
       att:1,
       auto_step:1,
       auto_mode:1,
       unknown11:2,
       bell:1;
    } memory[%d];                        // DN, VW, FM, AM
                                        // AMS DN, AMS VW, AMS FM

    #seekto 0x280A;
    struct {
    u8 nosubvfo:1,
    unknown:3,
    pskip:1,                            // PSkip (Select?)
    skip:1,                             // Skip memory during scan
    used:1,                             // Memory used
    valid:1;                            // Always 1?
    } flag[%d];
    """

MEM_CALLSIGN_FORMAT = """
#seekto 0x0ced0;
    struct {
    char callsign[10];              // 63 MYCALL    Set the call sign. (up to 10 characters)
    u16 charset;                    // character set ID
    } my_call;
    """

MEM_CHECKSUM_FORMAT = """
    #seekto 0xFECA;
    u8 checksum;
    """

TMODES = ["", "Tone", "TSQL", "DTCS"]
DUPLEX = ["", "-", "+", "split"]

MODES = ["FM", "AM", "NFM"]

STEPS = [0, 5, 6.25, 10, 12.5, 15, 20, 25, 50, 100]  # 0 = auto
RFSQUELCH = ["OFF", "S1", "S2", "S3", "S4", "S5", "S6", "S7", "S8"]

SKIPS = ["", "S", "P"]
FT70_DTMF_CHARS = list("0123456789ABCDEF-")

CHARSET = ["%i" % int(x) for x in range(0, 10)] + \
          [chr(x) for x in range(ord("A"), ord("Z") + 1)] + \
          [" ", ] + \
          [chr(x) for x in range(ord("a"), ord("z") + 1)] + \
          list(".,:;*#_-/&()@!?^ ") + list("\x00" * 100)

POWER_LEVELS = [chirp_common.PowerLevel("Hi", watts=5.00),
                chirp_common.PowerLevel("Mid", watts=2.00),
                chirp_common.PowerLevel("Low", watts=.50)]


class FT70Bank(chirp_common.NamedBank):
    """A FT70 bank"""

    def get_name(self):
        _bank = self._model._radio._memobj.bank_info[self.index]
        name = ""
        for i in _bank.name:
            if i == 0xff:
                break
            name += chr(i & 0xFF)
        return name.rstrip()

    def set_name(self, name):
        _bank = self._model._radio._memobj.bank_info[self.index]
        _bank.name = [ord(x) for x in name.ljust(6, chr(0xFF))[:6]]


class FT70BankModel(chirp_common.BankModel):
    """A FT70 bank model"""

    def __init__(self, radio, name='Banks'):
        super(FT70BankModel, self).__init__(radio, name)

        _banks = self._radio._memobj.bank_info
        self._bank_mappings = []
        for index, _bank in enumerate(_banks):
            bank = FT70Bank(self, "%i" % index, "BANK-%i" % index)
            bank.index = index
            self._bank_mappings.append(bank)

    def get_num_mappings(self):
        return len(self._bank_mappings)

    def get_mappings(self):
        return self._bank_mappings

    def _channel_numbers_in_bank(self, bank):
        _bank_used = self._radio._memobj.bank_used[bank.index]
        if _bank_used.in_use == 0xFFFF:
            return set()

        _members = self._radio._memobj.bank_members[bank.index]
        return set([int(ch) + 1 for ch in _members.channel if ch != 0xFFFF])

    def _update_bank_with_channel_numbers(self, bank, channels_in_bank):
        _members = self._radio._memobj.bank_members[bank.index]
        if len(channels_in_bank) > len(_members.channel):
            raise Exception("Too many entries in bank %d" % bank.index)

        empty = 0
        for index, channel_number in enumerate(sorted(channels_in_bank)):
            _members.channel[index] = channel_number - 1
            empty = index + 1
        for index in range(empty, len(_members.channel)):
            _members.channel[index] = 0xFFFF

    def add_memory_to_mapping(self, memory, bank):
        channels_in_bank = self._channel_numbers_in_bank(bank)
        channels_in_bank.add(memory.number)
        self._update_bank_with_channel_numbers(bank, channels_in_bank)

        _bank_used = self._radio._memobj.bank_used[bank.index]
        _bank_used.in_use = 0x06

    def remove_memory_from_mapping(self, memory, bank):
        channels_in_bank = self._channel_numbers_in_bank(bank)
        try:
            channels_in_bank.remove(memory.number)
        except KeyError:
            raise Exception("Memory %i is not in bank %s. Cannot remove" %
                            (memory.number, bank))
        self._update_bank_with_channel_numbers(bank, channels_in_bank)

        if not channels_in_bank:
            _bank_used = self._radio._memobj.bank_used[bank.index]
            _bank_used.in_use = 0xFFFF

    def get_mapping_memories(self, bank):
        memories = []
        for channel in self._channel_numbers_in_bank(bank):
            memories.append(self._radio.get_memory(channel))

        return memories

    def get_memory_mappings(self, memory):
        banks = []
        for bank in self.get_mappings():
            if memory.number in self._channel_numbers_in_bank(bank):
                banks.append(bank)

        return banks


@directory.register
class FT70Radio(yaesu_clone.YaesuCloneModeRadio):
    """Yaesu FT-70DE"""
    BAUD_RATE = 38400
    VENDOR = "Yaesu"
    MODEL = "FT-70D"
    FORMATS = [directory.register_format('FT-70D ADMS-10', '*.ft70d')]

    _model = b"AH51G"
    _adms_ext = '.ft70d'

    _memsize = 65227  # 65227 read from dump
    _block_lengths = [10, 65217]
    _block_size = 32
    _mem_params = (900,  # size of memories array
                   900,  # size of flags array
                   )

    _has_vibrate = False
    _has_af_dual = True

    _BEEP_SELECT = ("Off", "Key+Scan", "Key")
    _OPENING_MESSAGE = ("Off", "DC", "Message")
    _MIC_GAIN = ("Level 1", "Level 2", "Level 3", "Level 4", "Level 5", "Level 6", "Level 7", "Level 8", "Level 9")
    _AMS_TX_MODE = ("TX Auto", "TX DIGITAL", "TX FM")
    _VW_MODE = ("On", "Off")
    _DIG_POP_UP = ("Off", "2sec", "4sec", "6sec", "8sec", "10sec", "20sec", "30sec", "60sec", "Continuous")
    _STANDBY_BEEP = ("On", "Off")
    _SCAN_RESUME = ["%.1fs" % (0.5 * x) for x in range(4, 21)] + \
                   ["Busy", "Hold"]
    _SCAN_RESTART = ["%.1fs" % (0.1 * x) for x in range(1, 10)] + \
                    ["%.1fs" % (0.5 * x) for x in range(2, 21)]
    _LAMP_KEY = ["Key %d sec" % x
                 for x in range(2, 11)] + ["Continuous", "OFF"]
    _LCD_DIMMER = ["Level %d" % x for x in range(1, 7)]
    _TOT_TIME = ["Off"] + ["%.1f min" % (0.5 * x) for x in range(1, 21)]
    _OFF_ON = ("Off", "On")
    _ON_OFF = ("On", "Off")
    _DTMF_MODE = ("Manual", "Auto")
    _DTMF_SPEED = ("50ms", "100ms")
    _DTMF_DELAY = ("50ms", "250ms", "450ms", "750ms", "1000ms")
    _TEMP_CF = ("Centigrade", "Fahrenheit")
    _APO_SELECT = ("Off", "0.5H", "1.0H", "1.5H", "2.0H", "2.5H", "3.0H", "3.5H", "4.0H", "4.5H", "5.0H",
                   "5.5H", "6.0H", "6.5H", "7.0H", "7.5H", "8.0H", "8.5H", "9.0H", "9.5H", "10.0H", "10.5H",
                   "11.0H", "11.5H", "12.0H")
    _MONI_TCALL = ("Monitor", "Tone-CALL")
    _HOME_REV = ("Home", "Reverse")
    _LOCK = ("KEY", "DIAL", "Key+Dial", "PTT", "Key+PTT", "Dial+PTT", "ALL")
    _PTT_DELAY = ("Off", "20 ms", "50 ms", "100 ms", "200 ms")
    _BEEP_LEVEL = ("Level 1", "Level 2", "Level 3", "Level 4", "Level 5", "Level 6", "Level 7")
    _SET_MODE = ("Level 1", "Level 2", "Level 3", "Level 4", "Level 5", "Level 6", "Level 7")
    _RX_SAVE = ("OFF", "0.2s", ".3s", ".4s", ".5s", ".6s", ".7s", ".8s", ".9s", "1.0s", "1.5s",
                "2.0s", "2.5s", "3.0s", "3.5s", "4.0s", "4.5s", "5.0s", "5.5s", "6.0s", "6.5s", "7.0s",
                "7.5s", "8.0s", "8.5s", "9.0s", "10.0s", "15s", "20s", "25s", "30s", "35s", "40s", "45s", "50s", "55s",
                "60s")
    _VFO_MODE = ("ALL", "BAND")
    _VFO_SCAN_MODE = ("BAND", "ALL")
    _MEMORY_SCAN_MODE = ("BAND", "ALL")

    _VOLUME = ["Level %d" % x for x in range(0, 32)]
    _SQUELCH = ["Level %d" % x for x in range(0, 16)]

    _DG_ID = ["%d" % x for x in range(0, 100)]
    _GM_RING = ("OFF", "IN RING", "AlWAYS")
    _GM_INTERVAL = ("LONG", "NORMAL", "OFF")

    _MYCALL_CHR_SET = list(string.ascii_uppercase) + list(string.digits) + ['-', '/']

    @classmethod
    def match_model(cls, filedata, filename):
        if filename.endswith(cls._adms_ext):
            return True
        else:
            return super().match_model(filedata, filename)

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()

        rp.pre_download = _(
            "1. Turn radio on.\n"
            "2. Connect cable to DATA terminal.\n"
            "3. Unclip battery.\n"
            "4. Press and hold in the [AMS] key and power key while clipping"
            " \n in back battery the"
            "(\"ADMS\" will appear on the display).\n"
            "5. <b>After clicking OK</b>, press the [BAND] key.\n")
        rp.pre_upload = _(
            "1. Turn radio on.\n"
            "2. Connect cable to DATA terminal.\n"
            "3. Unclip battery.\n"
            "4. Press and hold in the [AMS] key and power key while clipping"
            " \n in back battery the"
            "(\"ADMS\" will appear on the display).\n"
            "5. Press the [MODE] key (\"-WAIT-\" will appear on the LCD)."
            " OK</b>\n click <b>Then")
        return rp

    def process_mmap(self):

        mem_format = MEM_SETTINGS_FORMAT + MEM_FORMAT + MEM_CALLSIGN_FORMAT + MEM_CHECKSUM_FORMAT

        self._memobj = bitwise.parse(mem_format % self._mem_params, self._mmap)

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_dtcs_polarity = False
        rf.valid_modes = list(MODES)
        rf.valid_tmodes = list(TMODES)
        rf.valid_duplexes = list(DUPLEX)
        rf.valid_tuning_steps = list(STEPS)
        rf.valid_bands = [(500000, 999900000)]
        rf.valid_skips = SKIPS
        rf.valid_power_levels = POWER_LEVELS
        rf.valid_characters = "".join(CHARSET)
        rf.valid_name_length = 6
        rf.memory_bounds = (1, 900)
        rf.can_odd_split = True
        rf.has_ctone = False
        rf.has_bank_names = True
        rf.has_settings = True
        return rf

    def get_raw_memory(self, number):
        return "\n".join([repr(self._memobj.memory[number - 1]),
                          repr(self._memobj.flag[number - 1])])

    def _checksums(self):
        return [yaesu_clone.YaesuChecksum(0x0000, 0xFEC9)]  # The whole file -2 bytes

    @staticmethod
    def _add_ff_pad(val, length):
        return val.ljust(length, "\xFF")[:length]

    @classmethod
    def _strip_ff_pads(cls, messages):
        result = []
        for msg_text in messages:
            result.append(str(msg_text).rstrip("\xFF"))
        return result

    def get_memory(self, number):
        flag = self._memobj.flag[number - 1]
        _mem = self._memobj.memory[number - 1]

        mem = chirp_common.Memory()
        mem.number = number
        if not flag.used:
            mem.empty = True
        if not flag.valid:
            mem.empty = True
            return mem
        mem.freq = chirp_common.fix_rounded_step(int(_mem.freq) * 1000)
        mem.offset = int(_mem.offset) * 1000
        mem.rtone = mem.ctone = chirp_common.TONES[_mem.tone]
        self._get_tmode(mem, _mem)
        mem.duplex = DUPLEX[_mem.duplex]
        if mem.duplex == "split":
            mem.offset = chirp_common.fix_rounded_step(mem.offset)
        mem.mode = self._decode_mode(_mem)
        mem.dtcs = chirp_common.DTCS_CODES[_mem.dcs]
        mem.tuning_step = STEPS[_mem.tune_step]
        mem.power = self._decode_power_level(_mem)
        mem.skip = flag.pskip and "P" or flag.skip and "S" or ""
        mem.name = self._decode_label(_mem)

        return mem

    def _decode_label(self, mem):
        return str(mem.label).rstrip("\xFF")

    def _encode_label(self, mem):
        return self._add_ff_pad(mem.name.rstrip(), 6)

    def _encode_charsetbits(self, mem):
        # We only speak English here in chirpville
        return [0x00, 0x00]

    def _decode_power_level(self, mem):  # 3 High 2 Mid 1 Low
        return POWER_LEVELS[3 - mem.power]

    def _encode_power_level(self, mem):
        return 3 - POWER_LEVELS.index(mem.power)

    def _decode_mode(self, mem):
        mode = MODES[mem.mode]
        if mode == 'FM' and int(mem.deviation):
            return 'NFM'
        else:
            return mode

    def _encode_mode(self, mem):
        mode = mem.mode
        if mode == 'NFM':
            # Narrow is handled by a separate flag
            mode = 'FM'
        return MODES.index(mode)

    def _get_tmode(self, mem, _mem):
        mem.tmode = TMODES[_mem.tone_mode]

    def _set_tmode(self, _mem, mem):
        _mem.tone_mode = TMODES.index(mem.tmode)

    def _set_mode(self, _mem, mem):
        _mem.deviation = mem.mode == 'NFM'
        _mem.mode = self._encode_mode(mem)

    def _debank(self, mem):
        bm = self.get_bank_model()
        for bank in bm.get_memory_mappings(mem):
            bm.remove_memory_from_mapping(mem, bank)

    def set_memory(self, mem):
        _mem = self._memobj.memory[mem.number - 1]
        flag = self._memobj.flag[mem.number - 1]

        self._debank(mem)

        if not mem.empty and not flag.valid:
            self._wipe_memory(_mem)

        if mem.empty and flag.valid and not flag.used:
            flag.valid = False
            return
        flag.used = not mem.empty
        flag.valid = flag.used

        if mem.empty:
            return

        if mem.freq < 30000000 or \
                (mem.freq > 88000000 and mem.freq < 108000000) or \
                mem.freq > 580000000:
            flag.nosubvfo = True  # Masked from VFO B
        else:
            flag.nosubvfo = False  # Available in both VFOs

        _mem.freq = int(mem.freq / 1000)
        _mem.offset = int(mem.offset / 1000)
        _mem.tone = chirp_common.TONES.index(mem.rtone)
        self._set_tmode(_mem, mem)
        _mem.duplex = DUPLEX.index(mem.duplex)
        self._set_mode(_mem, mem)
        _mem.dcs = chirp_common.DTCS_CODES.index(mem.dtcs)
        _mem.tune_step = STEPS.index(mem.tuning_step)

        if mem.power:
            _mem.power = self._encode_power_level(mem)
        else:
            _mem.power = 3  # Set 3 - High power as the default

        _mem.label = self._encode_label(mem)
        charsetbits = self._encode_charsetbits(mem)
        _mem.charsetbits[0], _mem.charsetbits[1] = charsetbits

        flag.skip = mem.skip == "S"
        flag.pskip = mem.skip == "P"

        _mem.display_tag = 1    # Always Display Memory Name (For the moment..)

    @classmethod
    def _wipe_memory(cls, mem):
        mem.set_raw("\x00" * (mem.size() // 8))
        mem.unknown1 = 0x05

    def get_bank_model(self):
        return FT70BankModel(self)

    def _get_dtmf_settings(self):
        menu = RadioSettingGroup("dtmf_settings", "DTMF")
        dtmf = self._memobj.scan_settings

        val = RadioSettingValueList(
            self._DTMF_MODE,
            self._DTMF_MODE[dtmf.dtmf_mode])
        rs = RadioSetting("scan_settings.dtmf_mode", "DTMF Mode", val)
        menu.append(rs)

        val = RadioSettingValueList(
            self._DTMF_DELAY,
            self._DTMF_DELAY[dtmf.dtmf_delay])
        rs = RadioSetting(
            "scan_settings.dtmf_delay", "DTMF Delay", val)
        menu.append(rs)

        val = RadioSettingValueList(
            self._DTMF_SPEED,
            self._DTMF_SPEED[dtmf.dtmf_speed])
        rs = RadioSetting(
            "scan_settings.dtmf_speed", "DTMF Speed", val)
        menu.append(rs)

        for i in range(10):

            name = "dtmf_%02d" % (i + 1)
            if i == 9:
                name = "dtmf_%02d" % 0

            dtmfsetting = self._memobj.dtmf[i]
            dtmfstr = ""
            for c in dtmfsetting.memory:
                if c == 0xFF:
                    break
                if c < len(FT70_DTMF_CHARS):
                    dtmfstr += FT70_DTMF_CHARS[c]
            dtmfentry = RadioSettingValueString(0, 16, dtmfstr)
            dtmfentry.set_charset(
                FT70_DTMF_CHARS + list("abcdef "))  # Allow input in lowercase, space ? validation fails otherwise
            rs = RadioSetting(name, name.upper(), dtmfentry)
            rs.set_apply_callback(self.apply_dtmf, i)
            menu.append(rs)

        return menu

    def _get_display_settings(self):
        menu = RadioSettingGroup("display_settings", "Display")
        scan_settings = self._memobj.scan_settings

        val = RadioSettingValueList(
            self._LAMP_KEY,
            self._LAMP_KEY[scan_settings.lamp])
        rs = RadioSetting("scan_settings.lamp", "Lamp", val)
        menu.append(rs)

        val = RadioSettingValueList(
            self._LCD_DIMMER,
            self._LCD_DIMMER[scan_settings.lcd_dimmer])
        rs = RadioSetting("scan_settings.lcd_dimmer", "LCD Dimmer", val)
        menu.append(rs)

        opening_message = self._memobj.opening_message
        val = RadioSettingValueList(
            self._OPENING_MESSAGE,
            self._OPENING_MESSAGE[opening_message.flag])
        rs = RadioSetting("opening_message.flag", "Opening Msg Mode", val)
        menu.append(rs)

        return menu

    def _get_config_settings(self):
        menu = RadioSettingGroup("config_settings", "Config")
        scan_settings = self._memobj.scan_settings

        # 02 APO    Set the length of time until the transceiver turns off automatically.

        first_settings = self._memobj.first_settings
        val = RadioSettingValueList(
            self._APO_SELECT,
            self._APO_SELECT[first_settings.apo])
        rs = RadioSetting("first_settings.apo", "APO", val)
        menu.append(rs)

        # 03 BCLO   Turns the busy channel lockout function on/off.

        val = RadioSettingValueList(
            self._OFF_ON,
            self._OFF_ON[scan_settings.bclo])
        rs = RadioSetting("scan_settings.bclo", "Busy Channel Lockout", val)
        menu.append(rs)

        # 04 BEEP   Sets the beep sound function.

        beep_settings = self._memobj.beep_settings
        val = RadioSettingValueList(
            self._BEEP_SELECT,
            self._BEEP_SELECT[beep_settings.beep_select])
        rs = RadioSetting("beep_settings.beep_select", "Beep", val)
        menu.append(rs)

        # 05 BEP.LVL    Beep volume setting LEVEL1 - LEVEL4 - LEVEL7

        val = RadioSettingValueList(
            self._BEEP_LEVEL,
            self._BEEP_LEVEL[beep_settings.beep_level])
        rs = RadioSetting("beep_settings", "Beep Level", val)
        menu.append(rs)

        # 06 BEP.Edg    Sets the beep sound ON or OFF when a band edge is encountered.

        val = RadioSettingValueList(
            self._OFF_ON,
            self._OFF_ON[scan_settings.beep_edge])
        rs = RadioSetting("scan_settings.beep_edge", "Beep Band Edge", val)
        menu.append(rs)

        # 10 Bsy.LED    Turn the MODE/STATUS Indicator ON or OFF while receiving signals.

        val = RadioSettingValueList(
            self._ON_OFF,
            self._ON_OFF[scan_settings.busy_led])
        rs = RadioSetting("scan_settings.busy_led", "Busy LED", val)
        menu.append(rs)

        # 26 HOME/REV   Select the function of the [HOME/REV] key.

        val = RadioSettingValueList(
            self._HOME_REV,
            self._HOME_REV[scan_settings.home_rev])
        rs = RadioSetting("scan_settings.home_rev", "HOME/REV", val)
        menu.append(rs)

        # 27 HOME->VFO  Turn transfer VFO to the Home channel ON or OFF.

        val = RadioSettingValueList(
            self._OFF_ON,
            self._OFF_ON[scan_settings.home_vfo])
        rs = RadioSetting("scan_settings.home_vfo", "Home->VFO", val)
        menu.append(rs)

        # 30 LOCK       Configure the lock mode setting. KEY / DIAL / K+D / PTT / K+P / D+P / ALL

        val = RadioSettingValueList(
            self._LOCK,
            self._LOCK[scan_settings.lock])
        rs = RadioSetting("scan_settings.lock", "Lock Mode", val)
        menu.append(rs)

        # 32 Mon/T-Call Select the function of the [MONI/T-CALL] switch.

        val = RadioSettingValueList(
            self._MONI_TCALL,
            self._MONI_TCALL[scan_settings.moni])
        rs = RadioSetting("scan_settings.moni", "MONI/T-CALL", val)
        menu.append(rs)

        # 42 PTT.DLY    Set the PTT delay time. OFF / 20 ms / 50 ms / 100 ms / 200 ms

        val = RadioSettingValueList(
            self._PTT_DELAY,
            self._PTT_DELAY[scan_settings.ptt_delay])
        rs = RadioSetting("scan_settings.ptt_delay", "PTT Delay", val)
        menu.append(rs)

        # 45 RPT.ARS    Turn the ARS function on/off.

        val = RadioSettingValueList(
            self._OFF_ON,
            self._OFF_ON[scan_settings.ars])
        rs = RadioSetting("scan_settings.ars", "ARS", val)
        menu.append(rs)

        # 48 RX.SAVE    Set the battery save time. OFF / 0.2 s - 60.0 s

        val = RadioSettingValueList(
            self._RX_SAVE,
            self._RX_SAVE[scan_settings.rx_save])
        rs = RadioSetting("scan_settings.rx_save", "RX SAVE", val)
        menu.append(rs)

        # 60 VFO.MOD    Set the frequency setting range in the VFO mode by DIAL knob. ALL / BAND

        val = RadioSettingValueList(
            self._VFO_MODE,
            self._VFO_MODE[scan_settings.vfo_mode])
        rs = RadioSetting("scan_settings.vfo_mode", "VFO MODE", val)
        menu.append(rs)

        # 56 TOT        Set the timeout timer.

        val = RadioSettingValueList(
            self._TOT_TIME,
            self._TOT_TIME[scan_settings.tot])
        rs = RadioSetting("scan_settings.tot", "Transmit Timeout (TOT)", val)
        menu.append(rs)

        # 31 MCGAIN     Adjust the microphone gain level

        val = RadioSettingValueList(
            self._MIC_GAIN,
            self._MIC_GAIN[scan_settings.mic_gain])
        rs = RadioSetting("scan_settings.mic_gain", "Mic Gain", val)
        menu.append(rs)

        # VOLUME       Adjust the volume level

        scan_settings_2 = self._memobj.scan_settings_2
        val = RadioSettingValueList(
            self._VOLUME,
            self._VOLUME[scan_settings_2.volume])
        rs = RadioSetting("scan_settings_2.volume", "Volume", val)
        menu.append(rs)

        # Squelch       F key, Hold Monitor, Dial to adjust squelch level

        squelch_settings = self._memobj.squelch_settings
        val = RadioSettingValueList(
            self._SQUELCH,
            self._SQUELCH[squelch_settings.squelch])
        rs = RadioSetting("squelch_settings.squelch", "Squelch", val)
        menu.append(rs)

        return menu

    def _get_digital_settings(self):
        menu = RadioSettingGroup("digital_settings", "Digital")

        # MYCALL
        mycall = self._memobj.my_call
        mycallstr = str(mycall.callsign).rstrip("\xFF")

        mycallentry = RadioSettingValueString(0, 10, mycallstr, False, charset=self._MYCALL_CHR_SET)
        rs = RadioSetting('mycall.callsign', 'MYCALL', mycallentry)
        rs.set_apply_callback(self.apply_mycall, mycall)
        menu.append(rs)

        # Short Press AMS button AMS TX Mode

        digital_settings = self._memobj.digital_settings
        val = RadioSettingValueList(
            self._AMS_TX_MODE,
            self._AMS_TX_MODE[digital_settings.ams_tx_mode])
        rs = RadioSetting("digital_settings.ams_tx_mode", "AMS TX Mode", val)
        menu.append(rs)

        # 16 DIG VW  Turn the VW mode selection ON or OFF.

        val = RadioSettingValueList(
            self._VW_MODE,
            self._VW_MODE[digital_settings.vw_mode])
        rs = RadioSetting("digital_settings.vw_mode", "VW Mode", val)
        menu.append(rs)

        # TX DG-ID Long Press Mode Key, Dial

        val = RadioSettingValueList(
            self._DG_ID,
            self._DG_ID[digital_settings.tx_dg_id])
        rs = RadioSetting("digital_settings.tx_dg_id", "TX DG-ID", val)
        menu.append(rs)

        # RX DG-ID Long Press Mode Key, Mode Key to select, Dial

        val = RadioSettingValueList(
            self._DG_ID,
            self._DG_ID[digital_settings.rx_dg_id])
        rs = RadioSetting("digital_settings.rx_dg_id", "RX DG-ID", val)
        menu.append(rs)

        # 15 DIG.POP    Call sign display pop up time

        # 00 OFF     00
        # 0A 2s      10
        # 0B 4s      11
        # 0C 6s      12
        # 0D 8s      13
        # 0E 10s     14
        # 0F 20s     15
        # 10 30s     16
        # 11 60s     17
        # 12 CONT    18

        digital_settings_more = self._memobj.digital_settings_more

        val = RadioSettingValueList(
            self._DIG_POP_UP,
            self._DIG_POP_UP[
                0 if digital_settings_more.digital_popup == 0 else digital_settings_more.digital_popup - 9])

        rs = RadioSetting("digital_settings_more.digital_popup", "Digital Popup", val)
        rs.set_apply_callback(self.apply_digital_popup, digital_settings_more)
        menu.append(rs)

        # 07  BEP.STB    Standby Beep in the digital C4FM mode. On/Off

        val = RadioSettingValueList(
            self._STANDBY_BEEP,
            self._STANDBY_BEEP[digital_settings.standby_beep])
        rs = RadioSetting("digital_settings.standby_beep", "Standby Beep", val)
        menu.append(rs)

        return menu

    def _get_gm_settings(self):
        menu = RadioSettingGroup("first_settings", "Group Monitor")

        # 24 GM RNG Select the beep option while receiving digital GM information. OFF / IN RNG /ALWAYS

        first_settings = self._memobj.first_settings
        val = RadioSettingValueList(
            self._GM_RING,
            self._GM_RING[first_settings.gm_ring])
        rs = RadioSetting("first_settings.gm_ring", "GM Ring", val)
        menu.append(rs)

        # 25 GM INT Set the transmission interval of digital GM information. OFF / NORMAL / LONG

        scan_settings = self._memobj.scan_settings
        val = RadioSettingValueList(
            self._GM_INTERVAL,
            self._GM_INTERVAL[scan_settings.gm_interval])
        rs = RadioSetting("scan_settings.gm_interval", "GM Interval", val)
        menu.append(rs)

        return menu

    def _get_scan_settings(self):
        menu = RadioSettingGroup("scan_settings", "Scan")
        scan_settings = self._memobj.scan_settings

        # 23 DW RVT     Turn the "Priority Channel Revert" feature ON or OFF during Dual Receive.

        val = RadioSettingValueList(
            self._OFF_ON,
            self._OFF_ON[scan_settings.dw_rt])
        rs = RadioSetting("scan_settings.dw_rt", "Dual Watch Priority Channel Revert", val)
        menu.append(rs)

        # 21 DW INT Set the priority memory channel monitoring interval during Dual Receive. 0.1S - 5.0S - 10.0S

        val = RadioSettingValueList(
            self._SCAN_RESTART,
            self._SCAN_RESTART[scan_settings.dw_interval])
        rs = RadioSetting("scan_settings.dw_interval", "Dual Watch Interval", val)
        menu.append(rs)

        # 22 DW RSM Configure the scan stop mode settings for Dual Receive. 2.0S - 10.0 S / BUSY / HOLD

        first_settings = self._memobj.first_settings
        val = RadioSettingValueList(
            self._SCAN_RESUME,
            self._SCAN_RESUME[first_settings.dw_resume_interval])
        rs = RadioSetting("first_settings.dw_resume_interval", "Dual Watch Resume Interval", val)
        menu.append(rs)

        # 51 SCN.LMP   Set the scan lamp ON or OFF when scanning stops. OFF / ON

        val = RadioSettingValueList(
            self._OFF_ON,
            self._OFF_ON[scan_settings.scan_lamp])
        rs = RadioSetting("scan_settings.scan_lamp", "Scan Lamp", val)
        menu.append(rs)

        # 53 SCN.STR   Set the scanning restart time.  0.1 S - 2.0 S - 10.0 S

        val = RadioSettingValueList(
            self._SCAN_RESTART,
            self._SCAN_RESTART[scan_settings.scan_restart])
        rs = RadioSetting("scan_settings.scan_restart", "Scan Restart", val)
        menu.append(rs)

        # Scan Width Section

        # 50 SCV.WTH Set the VFO scan frequency range. BAND / ALL  - NOT FOUND!

        # Scan Resume Section

        # 52 SCN.RSM    Configure the scan stop mode settings. 2.0 S - 5.0 S - 10.0 S / BUSY / HOLD

        first_settings = self._memobj.first_settings
        val = RadioSettingValueList(
            self._SCAN_RESUME,
            self._SCAN_RESUME[first_settings.scan_resume])
        rs = RadioSetting("first_settings.scan_resume", "Scan Resume", val)
        menu.append(rs)

        return menu

    def _get_settings(self):
        top = RadioSettings(
            self._get_config_settings(),
            self._get_digital_settings(),
            self._get_display_settings(),
            self._get_dtmf_settings(),
            self._get_gm_settings(),
            self._get_scan_settings()
        )
        return top

    def get_settings(self):
        try:
            return self._get_settings()
        except:
            import traceback
            LOG.error("Failed to parse settings: %s", traceback.format_exc())
            return None

    @classmethod
    def apply_ff_padded_string(cls, setting, obj):
        setattr(obj, "padded_string", cls._add_ff_pad(setting.value.get_value().rstrip(), 6))

    def set_settings(self, settings):
        _mem = self._memobj
        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            if not element.changed():
                continue
            try:
                if element.has_apply_callback():
                    LOG.debug("Using apply callback")
                    try:
                        element.run_apply_callback()
                    except NotImplementedError as e:
                        LOG.error(e)
                    continue

                # Find the object containing setting.
                obj = _mem
                bits = element.get_name().split(".")
                setting = bits[-1]
                for name in bits[:-1]:
                    if name.endswith("]"):
                        name, index = name.split("[")
                        index = int(index[:-1])
                        obj = getattr(obj, name)[index]
                    else:
                        obj = getattr(obj, name)

                try:
                    old_val = getattr(obj, setting)
                    LOG.debug("Setting %s(%r) <= %s" % (
                        element.get_name(), old_val, element.value))
                    setattr(obj, setting, element.value)
                except AttributeError as e:
                    LOG.error("Setting %s is not in the memory map: %s" %
                              (element.get_name(), e))
            except Exception:
                LOG.debug(element.get_name())
                raise

    def apply_volume(cls, setting, vfo):
        val = setting.value.get_value()
        cls._memobj.vfo_info[(vfo * 2)].volume = val
        cls._memobj.vfo_info[(vfo * 2) + 1].volume = val

    def apply_dtmf(cls, setting, i):
        rawval = setting.value.get_value().upper().rstrip()
        val = [FT70_DTMF_CHARS.index(x) for x in rawval]
        for x in range(len(val), 16):
            val.append(0xFF)
        cls._memobj.dtmf[i].memory = val

    def apply_digital_popup(cls, setting, obj):
        rawval = setting.value.get_value()
        val = 0 if cls._DIG_POP_UP.index(rawval) == 0 else cls._DIG_POP_UP.index(rawval) + 9
        obj.digital_popup = val

    def apply_mycall(cls, setting, obj):
        cs = setting.value.get_value()
        if cs[0] in ('-', '/'):
            raise InvalidValueError("First character of call sign can't be - or /:  {0:s}".format(cs))
        else:
            obj.callsign = cls._add_ff_pad(cs.rstrip(), 10)

    def load_mmap(self, filename):
        if filename.lower().endswith(self._adms_ext):
            with open(filename, 'rb') as f:
                self._adms_header = f.read(0xED)
                if b'=ADMS10, Version=1.0.1.0' not in self._adms_header:
                    raise errors.ImageDetectFailed(
                        'Unsupported version found in ADMS file')
                LOG.debug('ADMS Header:\n%s',
                          util.hexprint(self._adms_header))
                self._mmap = memmap.MemoryMapBytes(self._model + f.read())
                LOG.info('Loaded ADMS file')
            self.process_mmap()
        else:
            chirp_common.CloneModeRadio.load_mmap(self, filename)

    def save_mmap(self, filename):
        if filename.lower().endswith(self._adms_ext):
            if not hasattr(self, '_adms_header'):
                raise Exception('Unable to save .img to %s' % self._adms_ext)
            with open(filename, 'wb') as f:
                f.write(self._adms_header)
                f.write(self._mmap.get_packed()[5:])
                LOG.info('Wrote file')
        else:
            chirp_common.CloneModeRadio.save_mmap(self, filename)



# Copyright 2008 Dan Smith <dsmith at danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from chirp.drivers import icf
from chirp import chirp_common, bitwise, errors, directory
from chirp.settings import RadioSetting, RadioSettingGroup, \
    RadioSettingValueList, RadioSettingValueBoolean, \
    RadioSettings

LOG = logging.getLogger(__name__)


ICV80_MEM_FORMAT = """
struct {
  ul16 freq;
  ul16 offset;
  char name[5];
  u8 unknown0:2,
     rtone:6;
  u8 unknown1:2,
     ctone:6;
  u8 unknown2:1,
     dtcs:7;
  u8 unknown3:5,
     tuning_step:3;
  u8 unknown4:2,
     mode:1,
     reverse_duplex:1,
     duplex:2,
     mult:2;
  u8 unknown6:2,
     dtcs_polarity:2,
     unknown8:2,
     tmode:2;
  u8 unknown9:5,
     tx_inhibit:1,
     power:2;
} memory[207];

u8 skip[32];

u8 unused[32];

#seekto 0x0e50;
struct {
  u8 unknown1:6,
     beep:2;
  u8 unknown2:3,
     tot:5;
  u8 unknown3;
  u8 unknown4:6,
     auto_pwr_off:2;
  u8 unknown5:6,
     lockout:2;
  u8 unknown6:7,
     squelch_delay:1;
  u8 unknown7_0;
  u8 unknown7_1:6,
     mem_display1:2;
  u8 unknown8:6,
     mem_display2:2; // CS stores the display value here as well
  u8 unknown9:7,
     dial_func:1;
  u8 unknown10:7,
     lcd:1;
  u8 unknown11:5,
     pwr_save:3;
  u8 unknown12:7,
     sel_speed:1;
  u8 unknown13:6,
     mic_mode:2;
  u8 unknown14:6,
     battery_save:2;
  u8 unknown15;
  u8 unknown16:6,
     resume:2;
  u8 unknown17:5,
     func_mode:3;
  u8 unknown18:6,
     backlight:2;
  u8 unknown19;
  u8 unknown:4,
     vox_gain:4;
  u8 unknown20:6,
     mic_gain:2;
  u8 unknown21:5,
     vox_delay:3;
  u8 unknown22:4,
     vox_tot:4;
  u8 unknown23[2];
  u8 unknown24:6,
     edge:2;
  u8 unknown25;
  u8 unknown26:7,
     auto_low_pwr:1;
  u8 unknown27[3];

} settings;


"""

SPECIAL_CHANNELS = {
    "1A": 200, "1B": 201,
    "2A": 202, "2B": 203,
    "3A": 204, "3B": 205,
    "C": 206,
}

TMODES = ["", "Tone", "TSQL", "DTCS"]
MODES = ["FM", "NFM"]
SKIPS = ["", "S"]
DUPLEXES = ["", "-", "+"]
DTCS_POLARITY = ["NN", "NR", "RN", "RR"]
TUNING_STEPS = [5.0, 10.0, 12.5, 15.0, 20.0, 25.0, 30.0, 50.0]
POWER_LEVELS = [
    chirp_common.PowerLevel("High", watts=5.5),
    chirp_common.PowerLevel("Low", watts=0.5),
    chirp_common.PowerLevel("Mid", watts=2.5)
]


@directory.register
class ICV80Radio(icf.IcomCloneModeRadio, chirp_common.ExperimentalRadio):
    """Icom IC-V80"""
    VENDOR = "Icom"
    MODEL = "IC-V80"

    _model = "\x32\x54\x00\x01"
    _memsize = 0x0E80
    _endframe = "Icom Inc\x2e7C"
    _can_hispeed = True
    _ranges = [(0x0000, 0x0CE0, 32),
               (0x0CE0, 0x0D40, 16),
               (0x0D40, 0x0E00, 32),
               (0x0E00, 0x0E20, 16),
               (0x0E20, 0x0E60, 32),
               (0x0E60, 0x0E70, 16),
               (0x0E70, 0x0E72,  2),
               (0x0E72, 0x0E77,  5),
               (0x0E77, 0x0E78,  1),
               (0x0E78, 0x0E80,  8),
               ]

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.experimental = ("This radio driver is currently under development, "
                           "and not all the features or functions may work as"
                           "expected. You should proceed with caution.")
        return rp

    def get_features(self):
        rf = chirp_common.RadioFeatures()

        rf.memory_bounds = (0, 199)
        rf.valid_modes = MODES
        rf.valid_tmodes = TMODES
        rf.valid_duplexes = DUPLEXES
        rf.valid_tuning_steps = TUNING_STEPS
        rf.valid_power_levels = POWER_LEVELS
        rf.valid_skips = SKIPS
        rf.valid_name_length = 5
        rf.valid_special_chans = sorted(SPECIAL_CHANNELS.keys())
        rf.valid_bands = [(136000000, 174000000)]
        rf.has_ctone = True
        rf.has_offset = True
        rf.has_bank = False
        rf.has_settings = True

        return rf

    def process_mmap(self):
        self._memobj = bitwise.parse(ICV80_MEM_FORMAT, self._mmap)

    def get_settings(self):
        _settings = self._memobj.settings

        setmode = RadioSettingGroup("setmode", "Set Mode")
        display = RadioSettingGroup("display", "Display")
        sounds = RadioSettingGroup("sounds", "Sounds")
        scan = RadioSettingGroup("scan", "Scan")
        settings = RadioSettings(setmode, display, sounds, scan)

        # TOT
        opts = ["Off"] + ["%d min" % t for t in range(1, 31)]
        setmode.append(
            RadioSetting(
                "tot", "Time out Timer",
                RadioSettingValueList(opts, opts[_settings.tot])))

        # Lockout
        opts = ["Off", "Rpt", "Busy"]
        setmode.append(
            RadioSetting(
                "lockout", "Lockout",
                RadioSettingValueList(opts, opts[_settings.lockout])))

        # Auto Power Off
        opts = ["Off", "30 min", "1 hr", "2 hrs"]
        setmode.append(
            RadioSetting(
                "auto_pwr_off", "Auto Power Off",
                RadioSettingValueList(opts, opts[_settings.auto_pwr_off])))

        # Power Save
        opts = ["Off", "1:2", "1:8", "1:16", "Auto"]
        setmode.append(
            RadioSetting(
                "pwr_save", "Power Save",
                RadioSettingValueList(opts, opts[_settings.pwr_save])))

        # Battery Save
        opts = ["Off", "Ni-MH", "Li-Ion"]
        setmode.append(
            RadioSetting(
                "battery_save", "Battery Save",
                RadioSettingValueList(opts, opts[_settings.battery_save])))

        # Auto Low Power
        opts = ["Off", "On"]
        setmode.append(
            RadioSetting(
                "auto_low_pwr", "Auto Low Power",
                RadioSettingValueList(opts, opts[_settings.auto_low_pwr])))

        # Squelch Delay
        opts = ["Short", "Long"]
        setmode.append(
            RadioSetting(
                "squelch_delay", "Squelch Delay",
                RadioSettingValueList(opts, opts[_settings.squelch_delay])))

        # MIC Simple Mode
        opts = ["Simple", "Normal 1", "Normal 2"]
        setmode.append(
            RadioSetting(
                "mic_mode", "Mic Simple Mode",
                RadioSettingValueList(opts, opts[_settings.mic_mode])))

        # MIC Gain
        opts = ["1", "2", "3", "4"]
        setmode.append(
            RadioSetting(
                "mic_gain", "Mic Gain",
                RadioSettingValueList(opts, opts[_settings.mic_gain])))

        # VOX Gain
        opts = ["Off"] + ["%d" % t for t in range(1, 11)]
        setmode.append(
            RadioSetting(
                "vox_gain", "VOX Gain",
                RadioSettingValueList(opts, opts[_settings.vox_gain])))

        # VOX Delay
        opts = ["0.5 sec", "1.0 sec", "1.5 sec", "2.0 sec", "2.5 sec",
                "3.0 sec"]
        setmode.append(
            RadioSetting(
                "vox_delay", "VOX Delay",
                RadioSettingValueList(opts, opts[_settings.vox_delay])))

        # VOX Time out Timer
        opts = ["Off", "1 min", "2 min", "3 min", "4 min", "5 min", "10 min",
                "15 min"]
        setmode.append(
            RadioSetting(
                "vox_tot", "VOX Time-Out Timer",
                RadioSettingValueList(opts, opts[_settings.vox_tot])))

        # Select Speed
        opts = ["Manual", "Auto"]
        setmode.append(
            RadioSetting(
                "sel_speed", "Select Speed",
                RadioSettingValueList(opts, opts[_settings.sel_speed])))

        # Dial Function
        opts = ["Audio Volume", "Tuning Dial"]
        setmode.append(
            RadioSetting(
                "dial_func", "Dial Function",
                RadioSettingValueList(opts, opts[_settings.dial_func])))

        # Function Mode
        opts = ["0 sec", "1 sec", "2 sec", "3 sec", "Manual"]
        setmode.append(
            RadioSetting(
                "func_mode", "Function Mode",
                RadioSettingValueList(opts, opts[_settings.func_mode])))

        # Backlight
        opts = ["Off", "On", "Auto"]
        display.append(
            RadioSetting(
                "backlight", "Backlight",
                RadioSettingValueList(opts, opts[_settings.backlight])))

        # LCD Contrast
        opts = ["Low", "Auto"]
        display.append(
            RadioSetting(
                "lcd", "LCD Contrast",
                RadioSettingValueList(opts, opts[_settings.lcd])))

        # Memory Display
        opts = ["Frequency", "Channel", "Name"]
        display.append(
            RadioSetting(
                "mem_display1", "Memory Display",
                RadioSettingValueList(opts, opts[_settings.mem_display1])))

        # Beep
        opts = ["Off", "1", "2", "3"]
        sounds.append(
            RadioSetting(
                "beep", "Beep",
                RadioSettingValueList(opts, opts[_settings.beep])))

        # Edge
        opts = ["All", "P1", "P2", "P3"]
        scan.append(
            RadioSetting(
                "edge", "Edge",
                RadioSettingValueList(opts, opts[_settings.edge])))

        # Resume
        opts = ["T-5", "T-10", "T-15", "P-2"]
        scan.append(
            RadioSetting(
                "resume", "Resume",
                RadioSettingValueList(opts, opts[_settings.resume])))

        return settings

    def set_settings(self, settings):
        _settings = self._memobj.settings
        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            if not element.changed():
                continue

            try:
                if element.has_apply_callback():
                    LOG.debug("Using apply callback")
                    element.run_apply_callback()
                else:
                    setting = element.get_name()
                    LOG.debug("Setting %s = %s" % (setting, element.value))
                    setattr(_settings, setting, element.value)
                    # This appears to need to be mirrored?
                    if element.get_name() == 'mem_display1':
                        _settings.mem_display2 = _settings.mem_display1
            except Exception:
                LOG.debug(element.get_name())
                raise

    def _get_memory(self, number, extd_number=None):
        bit = 1 << (number % 8)
        byte = number // 8

        _mem = self._memobj.memory[number]
        _unused = self._memobj.unused[byte]
        _skip = self._memobj.skip[byte]

        mem = chirp_common.Memory(number)

        if extd_number is not None:
            mem.extd_number = extd_number
            mem.immutable = ["name", "number", "extd_number", "skip"]
            if extd_number == "C":
                _unused = False

        if (_unused & bit):
            mem.empty = True
            return mem

        if int(_mem.mult):
            mult = 6250
        else:
            mult = 5000
        mem.freq = int(_mem.freq) * mult
        mem.offset = int(_mem.offset) * mult
        if mem.extd_number == "":
            mem.name = str(_mem.name).rstrip()
            mem.skip = (_skip & bit) and "S" or ""
        mem.duplex = DUPLEXES[_mem.duplex]
        mem.power = POWER_LEVELS[_mem.power]
        mem.tuning_step = TUNING_STEPS[_mem.tuning_step]
        mem.mode = MODES[_mem.mode]
        mem.tmode = TMODES[_mem.tmode]
        mem.rtone = chirp_common.TONES[_mem.rtone]
        mem.ctone = chirp_common.TONES[_mem.ctone]
        mem.dtcs = chirp_common.DTCS_CODES[_mem.dtcs]
        mem.dtcs_polarity = DTCS_POLARITY[_mem.dtcs_polarity]

        # Reverse duplex
        mem.extra = RadioSettingGroup("extra", "Extra")
        rev = RadioSetting("reverse_duplex", "Reverse duplex",
                           RadioSettingValueBoolean(bool(_mem.reverse_duplex)))
        rev.set_doc("Reverse duplex")
        mem.extra.append(rev)

        # Tx inhibit
        tx_inhibit = RadioSetting("tx_inhibit", "TX inhibit",
                                  RadioSettingValueBoolean(
                                      bool(not _mem.tx_inhibit)))
        tx_inhibit.set_doc("TX inhibit")
        mem.extra.append(tx_inhibit)

        return mem

    def get_memory(self, number):
        extd_number = None
        if isinstance(number, str):
            try:
                extd_number = number
                number = SPECIAL_CHANNELS[number]
            except KeyError:
                raise errors.InvalidMemoryLocation(
                    "Unknown channel %s" % number)

        return self._get_memory(number, extd_number)

    def _fill_memory(self, mem):
        number = mem.number
        _mem = self._memobj.memory[number]

        # zero-fill
        _mem.freq = 146010000 // 5000
        _mem.offset = 600000 // 5000
        _mem.name = str("").ljust(5)
        _mem.duplex = 0x0
        _mem.reverse_duplex = 0x0
        _mem.tx_inhibit = 0x1
        _mem.power = 0x0
        _mem.tuning_step = 0x0
        _mem.mode = 0x0
        _mem.tmode = 0x0
        _mem.rtone = 0x8
        _mem.ctone = 0x8
        _mem.dtcs = 0x0
        _mem.dtcs_polarity = 0x0

        for setting in mem.extra:
            setattr(_mem, setting.get_name(), setting.value)

    def set_memory(self, mem):
        bit = 1 << (mem.number % 8)
        byte = mem.number // 8

        _mem = self._memobj.memory[mem.number]
        _unused = self._memobj.unused[byte]
        _skip = (mem.extd_number == "") and self._memobj.skip[byte] or None
        assert (_mem)

        if mem.empty:
            self._fill_memory(mem)
            _unused |= bit
            if _skip is not None:
                _skip |= bit
            return

        _mem.set_raw(b'\x00' * 16)

        if chirp_common.required_step(mem.freq) == 12.5:
            mult = 6250
        else:
            mult = 5000
        _mem.mult = 0 if mult == 5000 else 3
        _mem.freq = mem.freq // mult
        _mem.offset = int(mem.offset) // mult
        _mem.name = str(mem.name).ljust(5)
        _mem.duplex = DUPLEXES.index(mem.duplex)
        power = mem.power or POWER_LEVELS[0]
        _mem.power = POWER_LEVELS.index(power)
        _mem.tuning_step = TUNING_STEPS.index(mem.tuning_step)
        _mem.mode = MODES.index(mem.mode)
        _mem.tmode = TMODES.index(mem.tmode)
        _mem.rtone = chirp_common.TONES.index(mem.rtone)
        _mem.ctone = chirp_common.TONES.index(mem.ctone)
        _mem.dtcs = chirp_common.DTCS_CODES.index(mem.dtcs)
        _mem.dtcs_polarity = DTCS_POLARITY.index(mem.dtcs_polarity)

        # Set used
        _unused &= ~bit

        # Set skip
        if _skip is not None:
            if mem.skip == "S":
                _skip |= bit
            else:
                _skip &= ~bit

        if mem.extra:
            _mem.tx_inhibit = not mem.extra['tx_inhibit'].value
            _mem.reverse_duplex = mem.extra['reverse_duplex'].value

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number])



# Copyright 2023:
# * Sander van der Wel, <svdwel@icloud.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from chirp.drivers import baofeng_common as bfc
from chirp import chirp_common, directory, memmap
from chirp import bitwise
from chirp.settings import RadioSetting, \
    RadioSettingValueBoolean, RadioSettingValueList, \
    InvalidValueError, RadioSettingValueString, \
    RadioSettings, RadioSettingGroup
import struct
from chirp import errors, util

LOG = logging.getLogger(__name__)

# Baofeng UV-17L magic string
MSTRING_UV17L = b"PROGRAMBFNORMALU"
MSTRING_UV17PROGPS = b"PROGRAMCOLORPROU"

DTMF_CHARS = "0123456789 *#ABCD"
STEPS = [2.5, 5.0, 6.25, 10.0, 12.5, 20.0, 25.0, 50.0]
LIST_STEPS = ["2.5", "5.0", "6.25", "10.0", "12.5", "20.0", "25.0", "50.0"]

LIST_AB = ["A", "B"]
LIST_BANDWIDTH = ["Wide", "Narrow"]
LIST_DTMFSPEED = ["%s ms" % x for x in [50, 100, 200, 300, 500]]
LIST_HANGUPTIME = ["%s s" % x for x in [3, 4, 5, 6, 7, 8, 9, 10]]
LIST_SIDE_TONE = ["Off", "Side Key", "ID", "Side + ID"]
LIST_PTTID = ["Off", "BOT", "EOT", "Both"]
LIST_TIMEOUT_ALARM = ["Off"] + ["%s sec" % x for x in range(1, 11)]
LIST_PILOT_TONE = ["1000 Hz", "1450 Hz", "1750 Hz", "2100 Hz"]
LIST_VOICE = ["Off", "English", "Chinese"]
LIST_WORKMODE = ["Frequency", "Channel"]
LIST_BEEP = ["Off", "Beep", "Voice", "Both"]
LIST_SCANMODE = ["Time", "Carrier", "Search"]
LIST_ALARMMODE = ["Local", "Send Tone", "Send Code"]
LIST_MENU_QUIT_TIME = ["%s sec" % x for x in range(5, 55, 5)] + ["60 sec"]
LIST_ID_DELAY = ["%s ms" % x for x in range(100, 3100, 100)]
LIST_QT_SAVEMODE = ["Both", "RX", "TX"]
LIST_SKEY2_SHORT = ["FM", "Scan", "Search", "Vox"]
LIST_RPT_TAIL_CLEAR = ["%s ms" % x for x in range(0, 1100, 100)]
LIST_VOX_DELAY_TIME = ["%s ms" % x for x in range(500, 2100, 100)]
LIST_VOX_LEVEL = ["Off"] + ["%s" % x for x in range(1, 10, 1)]
LIST_VOX_LEVEL_ALT = ["%s" % x for x in range(1, 10, 1)]
LIST_GPS_MODE = ["GPS", "Beidou", "GPS + Beidou"]
LIST_GPS_TIMEZONE = ["%s" % x for x in range(-12, 13, 1)]
LIST_SHIFTS = ["Off", "+", "-"]
CHARSET_GB2312 = chirp_common.CHARSET_ASCII
for x in range(0xB0, 0xD7):
    for y in range(0xA1, 0xFF):
        CHARSET_GB2312 += bytes([x, y]).decode('gb2312')


def model_match(cls, data):
    """Match the opened image to the correct version"""
    return data[cls.MEM_TOTAL:] == bytes(cls.MODEL, 'utf-8')


def _crypt(symbol_index, buffer):
    # Some weird encryption is used. From the table below, we only use "CO 7".
    tblEncrySymbol = [b"BHT ", b"CO 7", b"A ES", b" EIY", b"M PQ",
                      b"XN Y", b"RVB ", b" HQP", b"W RC", b"MS N",
                      b" SAT", b"K DH", b"ZO R", b"C SL", b"6RB ",
                      b" JCG", b"PN V", b"J PK", b"EK L", b"I LZ"]
    tbl_encrypt_symbols = tblEncrySymbol[symbol_index]
    dec_buffer = b""
    index1 = 0
    for index2 in range(len(buffer)):
        bool_encrypt_char = ((tbl_encrypt_symbols[index1] != 32) and
                             (buffer[index2] != 0) and
                             (buffer[index2] != 255) and
                             (buffer[index2] != tbl_encrypt_symbols[index1])
                             and (buffer[index2] !=
                                  (tbl_encrypt_symbols[index1] ^ 255)))
        if (bool_encrypt_char):
            dec_byte = buffer[index2] ^ tbl_encrypt_symbols[index1]
            dec_buffer += struct.pack('>B', dec_byte)
        else:
            dec_buffer += struct.pack('>B', buffer[index2])
        index1 = (index1 + 1) % 4
    return dec_buffer


def _sendmagic(radio, magic, response_len):
    bfc._rawsend(radio, magic)
    return bfc._rawrecv(radio, response_len)


def _do_ident(radio):
    # Flush input buffer
    bfc._clean_buffer(radio)

    # Ident radio
    ack = _sendmagic(radio, radio._magic, len(radio._fingerprint))

    if not ack.startswith(radio._fingerprint):
        if ack:
            LOG.debug(repr(ack))
        raise errors.RadioError("Radio did not respond as expected (A)")

    for magic, resplen in radio._magics:
        _sendmagic(radio, magic, resplen)

    return True


def _download(radio):
    """Get the memory map"""

    # Put radio in program mode and identify it
    _do_ident(radio)
    data = b""

    # UI progress
    status = chirp_common.Status()
    status.cur = 0
    status.max = radio.MEM_TOTAL // radio.BLOCK_SIZE
    status.msg = "Cloning from radio..."
    radio.status_fn(status)

    for i in range(len(radio.MEM_SIZES)):
        MEM_SIZE = radio.MEM_SIZES[i]
        MEM_START = radio.MEM_STARTS[i]
        for addr in range(MEM_START, MEM_START + MEM_SIZE,
                          radio.BLOCK_SIZE):
            frame = radio._make_read_frame(addr, radio.BLOCK_SIZE)
            # DEBUG
            LOG.debug("Frame=" + util.hexprint(frame))

            # Sending the read request
            bfc._rawsend(radio, frame)

            # Now we read data
            d = bfc._rawrecv(radio, radio.BLOCK_SIZE + 4)

            LOG.debug("Response Data= " + util.hexprint(d))
            d = _crypt(1, d[4:])

            # Aggregate the data
            data += d

            # UI Update
            status.cur = len(data) // radio.BLOCK_SIZE
            status.msg = "Cloning from radio..."
            radio.status_fn(status)
    return data


def _upload(radio):
    # Put radio in program mode and identify it
    _do_ident(radio)
    data = b""

    # UI progress
    status = chirp_common.Status()
    status.cur = 0
    status.max = radio.MEM_TOTAL // radio.BLOCK_SIZE
    status.msg = "Cloning from radio..."
    radio.status_fn(status)

    data_addr = 0x00
    radio_mem = radio.get_mmap()
    for i in range(len(radio.MEM_SIZES)):
        MEM_SIZE = radio.MEM_SIZES[i]
        MEM_START = radio.MEM_STARTS[i]
        for addr in range(MEM_START, MEM_START + MEM_SIZE,
                          radio.BLOCK_SIZE):
            data = radio_mem[data_addr:data_addr + radio.BLOCK_SIZE]
            data = _crypt(1, data)
            data_addr += radio.BLOCK_SIZE

            frame = radio._make_frame(b"W", addr, radio.BLOCK_SIZE, data)
            # DEBUG
            LOG.debug("Frame=" + util.hexprint(frame))

            # Sending the read request
            bfc._rawsend(radio, frame)

            # receiving the response
            ack = bfc._rawrecv(radio, 1)
            if ack != b"\x06":
                msg = "Bad ack writing block 0x%04x" % addr
                raise errors.RadioError(msg)

            # UI Update
            status.cur = data_addr // radio.BLOCK_SIZE
            status.msg = "Cloning to radio..."
            radio.status_fn(status)
    return data


@directory.register
class UV17Pro(bfc.BaofengCommonHT):
    """Baofeng UV-17Pro"""
    VENDOR = "Baofeng"
    MODEL = "UV-17Pro"
    NEEDS_COMPAT_SERIAL = False

    MEM_STARTS = [0x0000, 0x9000, 0xA000, 0xD000]
    MEM_SIZES = [0x8040, 0x0040, 0x02C0, 0x0040]

    MEM_TOTAL = 0x8380
    BLOCK_SIZE = 0x40
    BAUD_RATE = 115200

    download_function = _download
    upload_function = _upload

    _gmrs = False
    _bw_shift = False
    _has_support_for_banknames = False
    _has_workmode_support = True
    _has_savemode = True

    _tri_band = True
    _fileid = []
    _magic = MSTRING_UV17L
    _fingerprint = b"\x06"
    _magics = [(b"\x46", 16),
               (b"\x4d", 15),
               (b"\x53\x45\x4E\x44\x21\x05\x0D\x01\x01\x01\x04\x11\x08\x05" +
                b"\x0D\x0D\x01\x11\x0F\x09\x12\x09\x10\x04\x00", 1)]
    _fw_ver_start = 0x1EF0
    _recv_block_size = 0x40
    _mem_size = MEM_TOTAL
    _ack_block = True
    _has_when_to_send_aniid = True
    _vfoscan = False
    _has_gps = False
    _has_voxsw = False
    _has_pilot_tone = False
    _has_send_id_delay = False
    _has_skey2_short = False
    _scode_offset = 0

    MODES = ["NFM", "FM"]
    VALID_CHARS = chirp_common.CHARSET_ALPHANUMERIC + \
        "!@#$%^&*()+-=[]:\";'<>?,./"
    LENGTH_NAME = 12
    SKIP_VALUES = ["", "S"]
    DTCS_CODES = tuple(sorted(chirp_common.DTCS_CODES + (645,)))
    RXTX_CODES = ('Off', )
    LIST_PW_SAVEMODE = ["Off", "On"]
    for code in chirp_common.TONES:
        RXTX_CODES = (RXTX_CODES + (str(code), ))
    for code in DTCS_CODES:
        RXTX_CODES = (RXTX_CODES + ('D' + str(code) + 'N', ))
    for code in DTCS_CODES:
        RXTX_CODES = (RXTX_CODES + ('D' + str(code) + 'I', ))
    POWER_LEVELS = [chirp_common.PowerLevel("High", watts=5.00),
                    chirp_common.PowerLevel("Low",  watts=1.00)]
    _airband = (108000000, 136000000)
    _vhf_range = (136000000, 174000000)
    _vhf2_range = (200000000, 260000000)
    _uhf_range = (400000000, 520000000)
    _uhf2_range = (350000000, 390000000)

    VALID_BANDS = [_vhf_range, _vhf2_range,
                   _uhf_range]
    PTTID_LIST = LIST_PTTID
    SCODE_LIST = ["%s" % x for x in range(1, 21)]
    SQUELCH_LIST = ["Off"] + list("12345")
    LIST_POWERON_DISPLAY_TYPE = ["LOGO", "BATT voltage"]
    LIST_TIMEOUT = ["Off"] + ["%s sec" % x for x in range(15, 195, 15)]
    LIST_VOICE = ["English", "Chinese"]
    LIST_BACKLIGHT_TIMER = ["Always On"] + ["%s sec"
                                            % x for x in range(5, 25, 5)]
    LIST_MODE = ["Name", "Frequency", "Channel Number"]

    CHANNELS = 1000

    MEM_FORMAT = """
    struct {
      lbcd rxfreq[4];
      lbcd txfreq[4];
      ul16 rxtone;
      ul16 txtone;
      u8 scode;
      u8 pttid;
      u8 lowpower;
      u8 unknown1:1,
         wide:1,
         sqmode:2,
         bcl:1,
         scan:1,
         unknown2:1,
         fhss:1;
      u8 unknown3;
      u8 unknown4;
      u8 unknown5;
      u8 unknown6;
      char name[12];
    } memory[1000];

    struct vfo_entry {
      u8 freq[8];
      ul16 rxtone;
      ul16 txtone;
      u8 unknown0;
      u8 bcl;
      u8 sftd:3,
         scode:5;
      u8 unknown1;
      u8 lowpower;
      u8 unknown2:1,
         wide:1,
         unknown3:5,
         fhss:1;
      u8 unknown4;
      u8 step;
      u8 offset[6];
      u8 unknown5[2];
      u8 sqmode;
      u8 unknown6[3];
      };

    #seekto 0x8000;
    struct {
      struct vfo_entry a;
      struct vfo_entry b;
    } vfo;

    struct {
      u8 squelch;
      u8 savemode;
      u8 vox;
      u8 backlight;
      u8 dualstandby;
      u8 tot;
      u8 beep;
      u8 voicesw;
      u8 voice;
      u8 sidetone;
      u8 scanmode;
      u8 pttid;
      u8 pttdly;
      u8 chadistype;
      u8 chbdistype;
      u8 bcl;
      u8 autolock;
      u8 alarmmode;
      u8 alarmtone;
      u8 unknown1;
      u8 tailclear;
      u8 rpttailclear;
      u8 rpttaildet;
      u8 roger;
      u8 unknown2;
      u8 fmenable;
      u8 chbworkmode:4,
         chaworkmode:4;
      u8 keylock;
      u8 powerondistype;
      u8 tone;
      u8 unknown4[2];
      u8 voxdlytime;
      u8 menuquittime;
      u8 unknown5[6];
      u8 totalarm;
      u8 unknown6[2];
      u8 ctsdcsscantype;
      ul16 vfoscanmin;
      ul16 vfoscanmax;
      u8 gpsw;
      u8 gpsmode;
      u8 unknown7[2];
      u8 key2short;
      u8 unknown8[2];
      u8 rstmenu;
      u8 unknown9;
      u8 hangup;
      u8 voxsw;
      u8 gpstimezone;
    } settings;

    #seekto 0x8080;
    struct {
      u8 code[5];
      u8 unknown[1];
      u8 unused1:6,
         aniid:2;
      u8 dtmfon;
      u8 dtmfoff;
    } ani;

    #seekto 0x80A0;
    struct {
      u8 code[5];
      u8 name[10];
      u8 unused;
    } pttid[20];

    #seekto 0x8280;
    struct {
      char name[16];
    } bank_name[10];
    """

    def _make_read_frame(self, addr, length):
        """Pack the info in the header format"""
        frame = self._make_frame(b"\x52", addr, length)
        # Return the data
        return frame

    def _make_frame(self, cmd, addr, length, data=""):
        """Pack the info in the header format"""
        frame = cmd + struct.pack(">i", addr)[2:] + struct.pack("b", length)
        # add the data if set
        if len(data) != 0:
            frame += data
        # return the data
        return frame

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.experimental = \
            ('This driver is a beta version.\n'
             '\n'
             'Please save an unedited copy of your first successful\n'
             'download to a CHIRP Radio Images(*.img) file.'
             )
        rp.pre_download = _(
            "Follow these instructions to download your info:\n"
            "1 - Turn off your radio\n"
            "2 - Connect your interface cable\n"
            "3 - Turn on your radio\n"
            "4 - Do the download of your radio data\n")
        rp.pre_upload = _(
            "Follow this instructions to upload your info:\n"
            "1 - Turn off your radio\n"
            "2 - Connect your interface cable\n"
            "3 - Turn on your radio\n"
            "4 - Do the upload of your radio data\n")
        return rp

    def process_mmap(self):
        """Process the mem map into the mem object"""
        # make lines shorter for style check.
        self._memobj = bitwise.parse(self.MEM_FORMAT, self._mmap)

    # DTMF settings
    def apply_code(self, setting, obj, length):
        code = []
        for j in range(0, length):
            try:
                code.append(DTMF_CHARS.index(str(setting.value)[j]))
            except IndexError:
                code.append(0xFF)
        obj.code = code

    def get_settings_common_dtmf(self, dtmfe, _mem):
        for i in range(0, len(self.SCODE_LIST)):
            _codeobj = self._memobj.pttid[i].code
            _code = "".join([
                DTMF_CHARS[x] for x in _codeobj if int(x) < 0x1F])
            val = RadioSettingValueString(0, 5, _code, False)
            val.set_charset(DTMF_CHARS)
            pttid = RadioSetting("pttid/%i.code" % i,
                                 "Signal Code %i" % (i + 1), val)
            pttid.set_apply_callback(self.apply_code, self._memobj.pttid[i], 5)
            dtmfe.append(pttid)

        _codeobj = self._memobj.ani.code
        _code = "".join([DTMF_CHARS[x] for x in _codeobj if int(x) < 0x1F])
        val = RadioSettingValueString(0, 5, _code, False)
        val.set_charset(DTMF_CHARS)
        rs = RadioSetting("ani.code", "ANI Code", val)
        rs.set_apply_callback(self.apply_code, self._memobj.ani, 5)
        dtmfe.append(rs)

    def get_settings_pro_dtmf(self, dtmfe, _mem):
        if _mem.ani.dtmfon > 0xC3:
            val = 0x03
        else:
            val = _mem.ani.dtmfon
        rs = RadioSetting("ani.dtmfon", "DTMF Speed (on)",
                          RadioSettingValueList(LIST_DTMFSPEED,
                                                LIST_DTMFSPEED[val]))
        dtmfe.append(rs)

        if _mem.ani.dtmfoff > 0xC3:
            val = 0x03
        else:
            val = _mem.ani.dtmfoff
        rs = RadioSetting("ani.dtmfoff", "DTMF Speed (off)",
                          RadioSettingValueList(LIST_DTMFSPEED,
                                                LIST_DTMFSPEED[val]))
        dtmfe.append(rs)

        if self._has_when_to_send_aniid:
            rs = RadioSetting("ani.aniid", "When to send ANI ID",
                              RadioSettingValueList(LIST_PTTID,
                                                    LIST_PTTID[
                                                        _mem.ani.aniid]))
            dtmfe.append(rs)

        if _mem.settings.hangup >= len(LIST_HANGUPTIME):
            val = 0
        else:
            val = _mem.settings.hangup
        rs = RadioSetting("settings.hangup", "Hang-up time",
                          RadioSettingValueList(LIST_HANGUPTIME,
                                                LIST_HANGUPTIME[
                                                    val]))
        dtmfe.append(rs)

    def get_settings_common_basic(self, basic, _mem):
        if _mem.settings.squelch >= len(self.SQUELCH_LIST):
            val = 0x00
        else:
            val = _mem.settings.squelch
        rs = RadioSetting("settings.squelch", "Squelch",
                          RadioSettingValueList(
                              self.SQUELCH_LIST, self.SQUELCH_LIST[val]))
        basic.append(rs)

        if _mem.settings.tot >= len(self.LIST_TIMEOUT):
            val = 0x03
        else:
            val = _mem.settings.tot
        rs = RadioSetting("settings.tot", "Timeout Timer",
                          RadioSettingValueList(
                              self.LIST_TIMEOUT, self.LIST_TIMEOUT[val]))
        basic.append(rs)

        rs = RadioSetting("settings.dualstandby", "Dual Watch",
                          RadioSettingValueBoolean(_mem.settings.dualstandby))
        basic.append(rs)

        if _mem.settings.powerondistype >= len(self.LIST_POWERON_DISPLAY_TYPE):
            val = 0x00
        else:
            rs = RadioSetting("settings.powerondistype",
                              "Power On Display Type",
                              RadioSettingValueList(
                                  self.LIST_POWERON_DISPLAY_TYPE,
                                  self.LIST_POWERON_DISPLAY_TYPE[
                                      _mem.settings.powerondistype]))
            basic.append(rs)

        if _mem.settings.voice >= len(self.LIST_VOICE):
            val = 0x01
        else:
            val = _mem.settings.voice
        rs = RadioSetting("settings.voice", "Voice Prompt",
                          RadioSettingValueList(
                              self.LIST_VOICE, self.LIST_VOICE[val]))
        basic.append(rs)

        rs = RadioSetting("settings.voicesw", "Enable Voice",
                          RadioSettingValueBoolean(_mem.settings.voicesw))
        basic.append(rs)

        if _mem.settings.backlight >= len(self.LIST_BACKLIGHT_TIMER):
            val = 0x00
        else:
            val = _mem.settings.backlight
        rs = RadioSetting("settings.backlight", "Backlight Timer",
                          RadioSettingValueList(
                              self.LIST_BACKLIGHT_TIMER,
                              self.LIST_BACKLIGHT_TIMER[val]))
        basic.append(rs)

        rs = RadioSetting("settings.autolock", "Key Auto Lock",
                          RadioSettingValueBoolean(_mem.settings.autolock))
        basic.append(rs)

        rs = RadioSetting("settings.beep", "Beep",
                          RadioSettingValueList(
                              LIST_BEEP, LIST_BEEP[_mem.settings.beep]))
        basic.append(rs)

        rs = RadioSetting("settings.roger", "Roger",
                          RadioSettingValueBoolean(_mem.settings.roger))
        basic.append(rs)

        rs = RadioSetting("settings.chadistype", "Channel A display type",
                          RadioSettingValueList(
                              self.LIST_MODE,
                              self.LIST_MODE[_mem.settings.chadistype]))
        basic.append(rs)

        rs = RadioSetting("settings.chbdistype", "Channel B display type",
                          RadioSettingValueList(
                              self.LIST_MODE,
                              self.LIST_MODE[_mem.settings.chbdistype]))
        basic.append(rs)

    def get_settings_pro_basic(self, basic, _mem):

        if _mem.settings.savemode > len(self.LIST_PW_SAVEMODE):
            val = 0x01  # assume values out of range are some form of "On"
        else:
            val = _mem.settings.savemode
        rs = RadioSetting("settings.savemode", "Save Mode",
                          RadioSettingValueList(self.LIST_PW_SAVEMODE,
                                                self.LIST_PW_SAVEMODE[val]))
        basic.append(rs)

        rs = RadioSetting("settings.totalarm", "Timeout Timer Alarm",
                          RadioSettingValueList(
                              LIST_TIMEOUT_ALARM,
                              LIST_TIMEOUT_ALARM[_mem.settings.totalarm]))
        basic.append(rs)

        if self._has_pilot_tone:
            rs = RadioSetting("settings.tone", "Pilot Tone",
                              RadioSettingValueList(
                                LIST_PILOT_TONE,
                                LIST_PILOT_TONE[_mem.settings.tone]))
            basic.append(rs)

        rs = RadioSetting("settings.sidetone", "Side Tone",
                          RadioSettingValueList(
                              LIST_SIDE_TONE,
                              LIST_SIDE_TONE[_mem.settings.sidetone]))
        basic.append(rs)

        rs = RadioSetting("settings.tailclear", "Tail Clear",
                          RadioSettingValueBoolean(_mem.settings.tailclear))
        basic.append(rs)

        rs = RadioSetting("settings.scanmode", "Scan Mode",
                          RadioSettingValueList(
                              LIST_SCANMODE,
                              LIST_SCANMODE[_mem.settings.scanmode]))
        basic.append(rs)

        rs = RadioSetting("settings.alarmmode", "Alarm Mode",
                          RadioSettingValueList(
                              LIST_ALARMMODE,
                              LIST_ALARMMODE[_mem.settings.alarmmode]))
        basic.append(rs)

        rs = RadioSetting("settings.alarmtone", "Sound Alarm",
                          RadioSettingValueBoolean(_mem.settings.alarmtone))
        basic.append(rs)

        rs = RadioSetting("settings.keylock", "Key Lock",
                          RadioSettingValueBoolean(_mem.settings.keylock))
        basic.append(rs)

        rs = RadioSetting("settings.menuquittime", "Menu Quit Timer",
                          RadioSettingValueList(
                              LIST_MENU_QUIT_TIME,
                              LIST_MENU_QUIT_TIME[
                                  _mem.settings.menuquittime]))
        basic.append(rs)

        if self._has_send_id_delay:
            rs = RadioSetting("settings.pttdly", "Send ID Delay",
                              RadioSettingValueList(
                                LIST_ID_DELAY,
                                LIST_ID_DELAY[_mem.settings.pttdly]))
            basic.append(rs)

        rs = RadioSetting("settings.ctsdcsscantype", "QT Save Mode",
                          RadioSettingValueList(
                              LIST_QT_SAVEMODE,
                              LIST_QT_SAVEMODE[_mem.settings.ctsdcsscantype]))
        basic.append(rs)

        def getKey2shortIndex(value):
            key_to_index = {0x07: 0,
                            0x1C: 1,
                            0x1D: 2,
                            0x2D: 3}
            return key_to_index.get(int(value), 0)

        def apply_Key2short(setting, obj):
            val = str(setting.value)
            key_to_index = {'FM': 0x07,
                            'Scan': 0x1C,
                            'Search': 0x1D,
                            'Vox': 0x2D}
            obj.key2short = key_to_index.get(val, 0x07)

        if self._has_skey2_short:
            rs = RadioSetting("settings.key2short", "Skey2 Short",
                              RadioSettingValueList(
                                LIST_SKEY2_SHORT,
                                LIST_SKEY2_SHORT[
                                    getKey2shortIndex(
                                        _mem.settings.key2short)]))
            rs.set_apply_callback(apply_Key2short, _mem.settings)
            basic.append(rs)

        rs = RadioSetting("settings.chaworkmode", "Channel A work mode",
                          RadioSettingValueList(
                              LIST_WORKMODE,
                              LIST_WORKMODE[_mem.settings.chaworkmode]))
        basic.append(rs)

        rs = RadioSetting("settings.chbworkmode", "Channel B work mode",
                          RadioSettingValueList(
                              LIST_WORKMODE,
                              LIST_WORKMODE[_mem.settings.chbworkmode]))
        basic.append(rs)

        rs = RadioSetting("settings.rpttailclear", "Rpt Tail Clear",
                          RadioSettingValueList(
                              LIST_RPT_TAIL_CLEAR,
                              LIST_RPT_TAIL_CLEAR[
                                  _mem.settings.rpttailclear]))
        basic.append(rs)

        rs = RadioSetting("settings.rpttaildet", "Rpt Tail Delay",
                          RadioSettingValueList(
                              LIST_RPT_TAIL_CLEAR,
                              LIST_RPT_TAIL_CLEAR[_mem.settings.rpttaildet]))
        basic.append(rs)

        rs = RadioSetting("settings.rstmenu", "Enable Menu Rst",
                          RadioSettingValueBoolean(_mem.settings.rstmenu))
        basic.append(rs)

        if self._has_voxsw:
            rs = RadioSetting("settings.voxsw", "Vox Switch",
                              RadioSettingValueBoolean(_mem.settings.voxsw))
            basic.append(rs)

            rs = RadioSetting("settings.vox", "Vox Level",
                              RadioSettingValueList(
                                LIST_VOX_LEVEL_ALT,
                                LIST_VOX_LEVEL_ALT[_mem.settings.vox]))
            basic.append(rs)
        else:
            rs = RadioSetting("settings.vox", "Vox Level",
                              RadioSettingValueList(
                                LIST_VOX_LEVEL,
                                LIST_VOX_LEVEL[_mem.settings.vox]))
            basic.append(rs)

        rs = RadioSetting("settings.voxdlytime", "Vox Delay Time",
                          RadioSettingValueList(
                              LIST_VOX_DELAY_TIME,
                              LIST_VOX_DELAY_TIME[_mem.settings.voxdlytime]))
        basic.append(rs)

        if self._has_gps:
            rs = RadioSetting("settings.gpsw", "GPS On",
                              RadioSettingValueBoolean(_mem.settings.gpsw))
            basic.append(rs)

            rs = RadioSetting("settings.gpsmode", "GPS Mode",
                              RadioSettingValueList(
                                LIST_GPS_MODE,
                                LIST_GPS_MODE[_mem.settings.gpsmode]))
            basic.append(rs)

            rs = RadioSetting("settings.gpstimezone", "GPS Timezone",
                              RadioSettingValueList(
                                LIST_GPS_TIMEZONE,
                                LIST_GPS_TIMEZONE[_mem.settings.gpstimezone]))
            basic.append(rs)

        rs = RadioSetting("settings.fmenable", "Disable FM radio",
                          RadioSettingValueBoolean(_mem.settings.fmenable))
        basic.append(rs)

    def get_settings_common_workmode(self, workmode, _mem):

        vfoA = RadioSettingGroup("vfoA", "VFO A")
        vfoB = RadioSettingGroup("vfoB", "VFO B")

        def my_validate(value):
            value = chirp_common.parse_freq(value)
            for band in self.VALID_BANDS:
                if value > band[0] and value < band[1]:
                    return chirp_common.format_freq(value)
            msg = ("{0} is not in a valid band.".format(value))
            raise InvalidValueError(msg)

        def apply_freq(setting, obj):
            value = chirp_common.parse_freq(str(setting.value)) / 10
            for i in range(7, -1, -1):
                obj.freq[i] = value % 10
                value /= 10

        freqA = RadioSettingValueString(0, 10,
                                        bfc.bcd_decode_freq(_mem.vfo.a.freq))
        freqA.set_validate_callback(my_validate)
        rs = RadioSetting("vfo.a.freq", "Frequency", freqA)
        rs.set_apply_callback(apply_freq, _mem.vfo.a)
        vfoA.append(rs)

        freqB = RadioSettingValueString(0, 10,
                                        bfc.bcd_decode_freq(_mem.vfo.b.freq))
        freqB.set_validate_callback(my_validate)
        rs = RadioSetting("vfo.b.freq", "Frequency", freqB)
        rs.set_apply_callback(apply_freq, _mem.vfo.b)
        vfoB.append(rs)

        if _mem.vfo.a.sftd >= len(LIST_SHIFTS):
            val = 0
        else:
            val = _mem.vfo.a.sftd
        rs = RadioSetting("vfo.a.sftd", "Shift",
                          RadioSettingValueList(LIST_SHIFTS, LIST_SHIFTS[val]))
        vfoA.append(rs)

        if _mem.vfo.b.sftd >= len(LIST_SHIFTS):
            val = 0
        else:
            val = _mem.vfo.b.sftd
        rs = RadioSetting("vfo.b.sftd", "Shift",
                          RadioSettingValueList(LIST_SHIFTS, LIST_SHIFTS[val]))
        vfoB.append(rs)

        def convert_bytes_to_offset(bytes):
            real_offset = 0
            for byte in bytes:
                real_offset = (real_offset * 10) + byte
            return chirp_common.format_freq(real_offset * 1000)

        def apply_offset(setting, obj):
            value = chirp_common.parse_freq(str(setting.value)) / 1000
            for i in range(5, -1, -1):
                obj.offset[i] = value % 10
                value /= 10

        offA = RadioSettingValueString(
                0, 10, convert_bytes_to_offset(_mem.vfo.a.offset))
        rs = RadioSetting("vfo.a.offset",
                          "Offset (0.0-999.999)", offA)
        rs.set_apply_callback(apply_offset, _mem.vfo.a)
        vfoA.append(rs)

        offB = RadioSettingValueString(
                0, 10, convert_bytes_to_offset(_mem.vfo.b.offset))
        rs = RadioSetting("vfo.b.offset",
                          "Offset (0.0-999.999)", offB)
        rs.set_apply_callback(apply_offset, _mem.vfo.b)
        vfoB.append(rs)

        POWER_LEVELS = [str(x) for x in self.POWER_LEVELS]
        if _mem.vfo.a.lowpower >= len(POWER_LEVELS):
            val = 0
        else:
            val = _mem.vfo.a.lowpower
        rs = RadioSetting("vfo.a.lowpower", "Power",
                          RadioSettingValueList(POWER_LEVELS,
                                                POWER_LEVELS[val]))
        vfoA.append(rs)

        if _mem.vfo.b.lowpower >= len(POWER_LEVELS):
            val = 0
        else:
            val = _mem.vfo.b.lowpower
        rs = RadioSetting("vfo.b.lowpower", "Power",
                          RadioSettingValueList(POWER_LEVELS,
                                                POWER_LEVELS[val]))
        vfoB.append(rs)

        if _mem.vfo.a.wide >= len(LIST_BANDWIDTH):
            val = 0
        else:
            val = _mem.vfo.a.wide
        rs = RadioSetting("vfo.a.wide", "Bandwidth",
                          RadioSettingValueList(LIST_BANDWIDTH,
                                                LIST_BANDWIDTH[val]))
        vfoA.append(rs)

        if _mem.vfo.b.wide >= len(LIST_BANDWIDTH):
            val = 0
        else:
            val = _mem.vfo.b.wide
        rs = RadioSetting("vfo.b.wide", "Bandwidth",
                          RadioSettingValueList(LIST_BANDWIDTH,
                                                LIST_BANDWIDTH[val]))
        vfoB.append(rs)

        if _mem.vfo.a.scode >= len(self.SCODE_LIST):
            val = 0
        else:
            val = _mem.vfo.a.scode
        rs = RadioSetting("vfo.a.scode", "Signal Code",
                          RadioSettingValueList(self.SCODE_LIST,
                                                self.SCODE_LIST[val]))
        vfoA.append(rs)

        if _mem.vfo.b.scode >= len(self.SCODE_LIST):
            val = 0
        else:
            val = _mem.vfo.b.scode
        rs = RadioSetting("vfo.b.scode", "Signal Code",
                          RadioSettingValueList(self.SCODE_LIST,
                                                self.SCODE_LIST[val]))
        vfoB.append(rs)

        if _mem.vfo.a.step >= len(STEPS):
            val = 0
        else:
            val = _mem.vfo.a.step
        rs = RadioSetting("vfo.a.step", "Tuning Step",
                          RadioSettingValueList(LIST_STEPS, LIST_STEPS[val]))
        vfoA.append(rs)

        if _mem.vfo.b.step >= len(STEPS):
            val = 0
        else:
            val = _mem.vfo.b.step
        rs = RadioSetting("vfo.b.step", "Tuning Step",
                          RadioSettingValueList(LIST_STEPS, LIST_STEPS[val]))
        vfoB.append(rs)

        workmode.append(vfoA)
        workmode.append(vfoB)

    def get_settings_common_bank(self, bank, _mem):

        def _filterName(name):
            fname = b""
            for char in name:
                if ord(str(char)) == 255:
                    break
                fname += int(char).to_bytes(1, 'big')
            return fname.decode('gb2312').strip()

        def apply_bankname(setting, obj):
            name = str(setting.value).encode('gb2312')[:16].ljust(16, b"\xff")
            obj.name = name

        if self._has_support_for_banknames:
            for i in range(0, 10):
                _nameobj = self._memobj.bank_name[i]
                rs = RadioSetting("bank_name/%i.name" % i,
                                  "Bank name %i" % (i + 1),
                                  RadioSettingValueString(
                                      0, 16, _filterName(_nameobj.name), False,
                                      CHARSET_GB2312))
                rs.set_apply_callback(apply_bankname, _nameobj)
                bank.append(rs)

    def get_settings(self):
        """Translate the bit in the mem_struct into settings in the UI"""
        _mem = self._memobj
        supported = []

        basic = RadioSettingGroup("basic", "Basic Settings")
        self.get_settings_common_basic(basic, _mem)
        self.get_settings_pro_basic(basic, _mem)
        supported.append(basic)  # add basic menu

        if self._has_workmode_support:
            workmode = RadioSettingGroup("workmode", "Work Mode Settings")
            self.get_settings_common_workmode(workmode, _mem)
            supported.append(workmode)  # add workmode menu if supported

        dtmfe = RadioSettingGroup("dtmfe", "DTMF Encode Settings")
        self.get_settings_common_dtmf(dtmfe, _mem)
        self.get_settings_pro_dtmf(dtmfe, _mem)
        supported.append(dtmfe)  # add dtmfe menu

        if self._has_support_for_banknames:
            bank = RadioSettingGroup("bank", "Bank names")
            self.get_settings_common_bank(bank, _mem)
            supported.append(bank)  # add bank menu if supported

        top = RadioSettings(*tuple(supported))
        return top

    def set_settings(self, settings):
        _settings = self._memobj.settings
        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            else:
                try:
                    name = element.get_name()
                    if "." in name:
                        bits = name.split(".")
                        obj = self._memobj
                        for bit in bits[:-1]:
                            if "/" in bit:
                                bit, index = bit.split("/", 1)
                                index = int(index)
                                obj = getattr(obj, bit)[index]
                            else:
                                obj = getattr(obj, bit)
                        setting = bits[-1]
                    else:
                        obj = _settings
                        setting = element.get_name()

                    if element.has_apply_callback():
                        LOG.debug("Using apply callback")
                        element.run_apply_callback()
                    elif element.value.get_mutable():
                        LOG.debug("Setting %s = %s" % (setting, element.value))
                        setattr(obj, setting, element.value)
                except Exception:
                    LOG.debug(element.get_name())
                    raise

    def sync_in(self):
        """Download from radio"""
        try:
            data = self.download_function()
        except errors.RadioError:
            # Pass through any real errors we raise
            raise
        except Exception:
            # If anything unexpected happens, make sure we raise
            # a RadioError and log the problem
            LOG.exception('Unexpected error during download')
            raise errors.RadioError('Unexpected error communicating '
                                    'with the radio')
        self._mmap = memmap.MemoryMapBytes(data)

        self.process_mmap()

    def sync_out(self):
        """Upload to radio"""
        try:
            self.upload_function()
        except errors.RadioError:
            raise
        except Exception:
            # If anything unexpected happens, make sure we raise
            # a RadioError and log the problem
            LOG.exception('Unexpected error during upload')
            raise errors.RadioError('Unexpected error communicating '
                                    'with the radio')

    def get_features(self):
        """Get the radio's features"""

        rf = chirp_common.RadioFeatures()
        rf.has_settings = True
        rf.has_bank = False
        rf.has_tuning_step = False
        rf.can_odd_split = True
        rf.has_name = True
        rf.has_offset = True
        rf.has_mode = True
        rf.has_dtcs = True
        rf.has_rx_dtcs = True
        rf.has_dtcs_polarity = True
        rf.has_ctone = True
        rf.has_cross = True
        rf.valid_modes = self.MODES
        rf.valid_characters = self.VALID_CHARS
        rf.valid_name_length = self.LENGTH_NAME
        if self._gmrs:
            rf.valid_duplexes = ["", "+", "off"]
        else:
            rf.valid_duplexes = ["", "-", "+", "split", "off"]
        rf.valid_tmodes = ['', 'Tone', 'TSQL', 'DTCS', 'Cross']
        rf.valid_cross_modes = [
            "Tone->Tone",
            "DTCS->",
            "->DTCS",
            "Tone->DTCS",
            "DTCS->Tone",
            "->Tone",
            "DTCS->DTCS"]
        rf.valid_skips = self.SKIP_VALUES
        rf.valid_dtcs_codes = self.DTCS_CODES
        rf.memory_bounds = (1, self.CHANNELS)
        rf.valid_power_levels = self.POWER_LEVELS
        rf.valid_bands = self.VALID_BANDS
        rf.valid_tuning_steps = STEPS

        return rf

    def decode_tone(self, val):
        mode = ""
        pol = "N"
        if val in [0, 0xFFFF]:
            xval = 0
        elif val >= 0x0258:
            mode = "Tone"
            xval = int(val) / 10.0
        elif val <= 0x0258:
            mode = "DTCS"
            if val > 0x69:
                index = val - 0x6A
                pol = "R"
            else:
                index = val - 1
            xval = self.DTCS_CODES[index]
        else:
            LOG.warn("Bug: tone is %04x" % val)
        return mode, xval, pol

    def encode_tone(self, memtone, mode, tone, pol):
        if mode == "Tone":
            memtone.set_value(int(tone * 10))
        elif mode == "TSQL":
            memtone.set_value(int(tone * 10))
        elif mode == "DTCS":
            if pol == 'R':
                memtone.set_value(self.DTCS_CODES.index(tone) + 1 + 0x69)
            else:
                memtone.set_value(self.DTCS_CODES.index(tone) + 1)
        else:
            memtone.set_value(0)

    def split_txfreq(self, _mem, freq):
        if self._is_txinh(_mem):
            # TX freq not set
            duplex = "off"
            offset = 0
        else:
            offset = (int(_mem.txfreq) * 10) - freq
            if offset != 0:
                if bfc._split(self.get_features(), freq, int(
                          _mem.txfreq) * 10):
                    duplex = "split"
                    offset = int(_mem.txfreq) * 10
                elif offset < 0:
                    offset = abs(offset)
                    duplex = "-"
                elif offset > 0:
                    duplex = "+"
            else:
                duplex = ""
                offset = 0
        return offset, duplex

    def get_memory_common(self, _mem, name, mem):
        if _mem.get_raw()[0] == 255:
            mem.empty = True
            return mem

        mem.freq = int(_mem.rxfreq) * 10

        # TX freq set
        mem.offset, mem.duplex = self.split_txfreq(_mem, mem.freq)

        txtone = self.decode_tone(_mem.txtone)
        rxtone = self.decode_tone(_mem.rxtone)
        chirp_common.split_tone_decode(mem, txtone, rxtone)

        if not _mem.scan:
            mem.skip = "S"

        levels = self.POWER_LEVELS
        try:
            mem.power = levels[_mem.lowpower]
        except IndexError:
            mem.power = levels[0]

        mem.mode = _mem.wide and self.MODES[0] or self.MODES[1]
        if (mem.freq >= self._airband[0] and mem.freq <= self._airband[1]):
            # NOTE: AM is not in valid_modes because you can't arbitrarily
            # enable it on this radio. However, we can expose it as immutable
            # which will display properly in the UI and not allow the user
            # to change those channels to FM.
            mem.mode = "AM"
            mem.immutable = ['mode']

        mem.extra = RadioSettingGroup("Extra", "extra")

        rs = RadioSetting("bcl", "BCL",
                          RadioSettingValueBoolean(_mem.bcl))
        mem.extra.append(rs)

        rs = RadioSetting("pttid", "PTT ID",
                          RadioSettingValueList(self.PTTID_LIST,
                                                self.PTTID_LIST[
                                                    _mem.pttid]))
        mem.extra.append(rs)

        scode = (_mem.scode - self._scode_offset) % len(self.SCODE_LIST)
        rs = RadioSetting("scode", "S-CODE",
                          RadioSettingValueList(self.SCODE_LIST,
                                                self.SCODE_LIST[
                                                    scode]))
        mem.extra.append(rs)

        mem.name = str(name).replace('\xFF', ' ').replace('\x00', ' ').rstrip()

    def get_raw_memory(self, number):
        return self._memobj.memory[number - 1]

    def get_memory(self, number):
        _mem = self.get_raw_memory(number)

        mem = chirp_common.Memory()
        mem.number = number

        self.get_memory_common(_mem, _mem.name, mem)

        return mem

    def unsplit_txfreq(self, mem):
        _mem = self.get_raw_memory(mem.number)
        if mem.duplex == "off":
            for i in range(0, 4):
                _mem.txfreq[i].set_raw(b"\xFF")
        elif mem.duplex == "split":
            _mem.txfreq = mem.offset / 10
        elif mem.duplex == "+":
            _mem.txfreq = (mem.freq + mem.offset) / 10
        elif mem.duplex == "-":
            _mem.txfreq = (mem.freq - mem.offset) / 10
        else:
            _mem.txfreq = mem.freq / 10

    def set_memory_common(self, mem, _mem):
        _mem.rxfreq = mem.freq / 10
        self.unsplit_txfreq(mem)

        ((txmode, txtone, txpol), (rxmode, rxtone, rxpol)) = \
            chirp_common.split_tone_encode(mem)
        self.encode_tone(_mem.txtone, txmode, txtone, txpol)
        self.encode_tone(_mem.rxtone, rxmode, rxtone, rxpol)

        _mem.scan = mem.skip != "S"
        _mem.wide = mem.mode == self.MODES[0]

        if mem.power:
            _mem.lowpower = self.POWER_LEVELS.index(mem.power)
        else:
            _mem.lowpower = 0

        # extra settings
        if len(mem.extra) > 0:
            # there are setting, parse
            for setting in mem.extra:
                if setting.get_name() == "scode":
                    setattr(_mem, setting.get_name(), str(int(setting.value) +
                                                          self._scode_offset))
                else:
                    setattr(_mem, setting.get_name(), setting.value)
        else:
            # there are no extra settings, load defaults
            _mem.bcl = 0
            _mem.pttid = 0
            _mem.scode = self._scode_offset

    def set_memory(self, mem):
        _mem = self.get_raw_memory(mem.number)

        _mem.set_raw(b"\x00"*16 + b"\xff" * 16)

        if mem.empty:
            _mem.set_raw(b"\xff" * 32)
            return

        _namelength = self.get_features().valid_name_length
        _mem.name = mem.name.ljust(_namelength, '\xFF')

        self.set_memory_common(mem, _mem)


@directory.register
class UV17ProGPS(UV17Pro):
    VENDOR = "Baofeng"
    MODEL = "UV-17ProGPS"

    _has_support_for_banknames = True
    _has_workmode_support = True
    _magic = MSTRING_UV17PROGPS
    _magics = [(b"\x46", 16),
               (b"\x4d", 7),
               (b"\x53\x45\x4E\x44\x21\x05\x0D\x01\x01" +
                b"\x01\x04\x11\x08\x05\x0D\x0D\x01\x11\x0F\x09\x12\x09" +
                b"\x10\x04\x00", 1)]
    _has_when_to_send_aniid = False
    _vfoscan = True
    _has_gps = True
    _has_voxsw = True
    _has_pilot_tone = True
    _has_send_id_delay = True
    _has_skey2_short = True
    VALID_BANDS = [UV17Pro._airband, UV17Pro._vhf_range, UV17Pro._vhf2_range,
                   UV17Pro._uhf_range, UV17Pro._uhf2_range]

    def check_set_memory_immutable_policy(self, existing, new):
        if (self._airband[0] <= new.freq <= self._airband[1] and
                new.mode == 'AM'):
            # This is valid, so mark mode as immutable so it doesn't get
            # blocked, and let the radio override it during set.
            new.immutable.append('mode')
            existing.immutable = []
        elif existing.mode == 'AM' and new.mode in self.MODES:
            # If we're going from a forced-AM channel to some valid one,
            # clear immutable so we allow the change.
            try:
                existing.immutable.remove('mode')
            except ValueError:
                pass
        super().check_set_memory_immutable_policy(existing, new)


@directory.register
class BF5RM(UV17Pro):
    VENDOR = "Baofeng"
    MODEL = "5RM"

    VALID_BANDS = [UV17Pro._airband, UV17Pro._vhf_range, UV17Pro._vhf2_range,
                   UV17Pro._uhf_range, UV17Pro._uhf2_range]
    POWER_LEVELS = [chirp_common.PowerLevel("High", watts=8.00),
                    chirp_common.PowerLevel("Low", watts=1.00),
                    chirp_common.PowerLevel("Medium", watts=5.00)]
    SCODE_LIST = ["%s" % x for x in range(1, 16)]
    LIST_PW_SAVEMODE = ["Off", "1:1", "1:2", "1:4"]
    _has_workmode_support = True

    def check_set_memory_immutable_policy(self, existing, new):
        if (self._airband[0] <= new.freq <= self._airband[1] and
                new.mode == 'AM'):
            # This is valid, so mark mode as immutable so it doesn't get
            # blocked, and let the radio override it during set.
            new.immutable.append('mode')
            existing.immutable = []
        elif existing.mode == 'AM' and new.mode in self.MODES:
            # If we're going from a forced-AM channel to some valid one,
            # clear immutable so we allow the change.
            try:
                existing.immutable.remove('mode')
            except ValueError:
                pass
        super().check_set_memory_immutable_policy(existing, new)



# Copyright 2019 Zhaofeng Li <hello@zhaofeng.li>
# Copyright 2013 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import division

import struct
import logging
from math import floor
from chirp import chirp_common, directory, bitwise, memmap, errors, util
from .uvb5 import BaofengUVB5

LOG = logging.getLogger(__name__)

mem_format = """
struct memory {
  lbcd freq[4];
  lbcd offset[4];
  u8 unknown1:2,
     txpol:1,
     rxpol:1,
     compander:1,
     scrambler:1,
     unknown2:2;
  u8 rxtoneb;
  u8 rxtonea;
  u8 txtoneb;
  u8 txtonea;
  u8 pttid:1,
     scanadd:1,
     isnarrow:1,
     bcl:1,
     highpower:1,
     revfreq:1,
     duplex:2;
  u8 unknown[2];
};

#seekto 0x0000;
char ident[32];
u8 blank[16];
struct memory vfo1;
struct memory channels[128];
#seekto 0x0840;
struct memory vfo3;
struct memory vfo2;

#seekto 0x09D0;
u16 fm_presets[16];

#seekto 0x0A30;
struct {
  u8 name[5];
} names[128];

#seekto 0x0D30;
struct {
  u8 squelch;
  u8 freqmode_ab:1,
     save_funct:1,
     backlight:1,
     beep_tone_disabled:1,
     roger:1,
     tdr:1,
     scantype:2;
  u8 language:1,
     workmode_b:1,
     workmode_a:1,
     workmode_fm:1,
     voice_prompt:1,
     fm:1,
     pttid:2;
  u8 unknown_0:5,
     timeout:3;
  u8 mdf_b:2,
     mdf_a:2,
     unknown_1:2,
     txtdr:2;
  u8 unknown_2:4,
     ste_disabled:1,
     unknown_3:2,
     sidetone:1;
  u8 vox;
  u8 unk1;
  u8 mem_chan_a;
  u16 fm_vfo;
  u8 unk4;
  u8 unk5;
  u8 mem_chan_b;
  u8 unk6;
  u8 last_menu; // number of last menu item accessed
} settings;

#seekto 0x0D50;
struct {
  u8 code[6];
} pttid;

#seekto 0x0F30;
struct {
  lbcd lower_vhf[2];
  lbcd upper_vhf[2];
  lbcd lower_uhf[2];
  lbcd upper_uhf[2];
} limits;

#seekto 0x0FF0;
struct {
  u8 vhfsquelch0;
  u8 vhfsquelch1;
  u8 vhfsquelch2;
  u8 vhfsquelch3;
  u8 vhfsquelch4;
  u8 vhfsquelch5;
  u8 vhfsquelch6;
  u8 vhfsquelch7;
  u8 vhfsquelch8;
  u8 vhfsquelch9;
  u8 unknown1[6];
  u8 uhfsquelch0;
  u8 uhfsquelch1;
  u8 uhfsquelch2;
  u8 uhfsquelch3;
  u8 uhfsquelch4;
  u8 uhfsquelch5;
  u8 uhfsquelch6;
  u8 uhfsquelch7;
  u8 uhfsquelch8;
  u8 uhfsquelch9;
  u8 unknown2[6];
  u8 vhfhipwr0;
  u8 vhfhipwr1;
  u8 vhfhipwr2;
  u8 vhfhipwr3;
  u8 vhfhipwr4;
  u8 vhfhipwr5;
  u8 vhfhipwr6;
  u8 vhfhipwr7;
  u8 vhflopwr0;
  u8 vhflopwr1;
  u8 vhflopwr2;
  u8 vhflopwr3;
  u8 vhflopwr4;
  u8 vhflopwr5;
  u8 vhflopwr6;
  u8 vhflopwr7;
  u8 uhfhipwr0;
  u8 uhfhipwr1;
  u8 uhfhipwr2;
  u8 uhfhipwr3;
  u8 uhfhipwr4;
  u8 uhfhipwr5;
  u8 uhfhipwr6;
  u8 uhfhipwr7;
  u8 uhflopwr0;
  u8 uhflopwr1;
  u8 uhflopwr2;
  u8 uhflopwr3;
  u8 uhflopwr4;
  u8 uhflopwr5;
  u8 uhflopwr6;
  u8 uhflopwr7;
} test;
"""


def do_ident(radio):
    radio.pipe.timeout = 3
    radio.pipe.write(b"\x05TROGRAM")
    for x in range(10):
        ack = radio.pipe.read(1)
        if ack == b'\x06':
            break
    else:
        raise errors.RadioError("Radio did not ack programming mode")
    radio.pipe.write(b"\x02")
    ident = radio.pipe.read(8)
    LOG.debug(util.hexprint(ident))
    if not ident.startswith(b'HKT511'):
        raise errors.RadioError("Unsupported model")
    radio.pipe.write(b"\x06")
    ack = radio.pipe.read(1)
    if ack != b"\x06":
        raise errors.RadioError("Radio did not ack ident")


def do_status(radio, direction, addr):
    status = chirp_common.Status()
    status.msg = "Cloning %s radio" % direction
    status.cur = addr
    status.max = 0x1000
    radio.status_fn(status)


def do_download(radio):
    do_ident(radio)
    data = b"TRI350 Radio Program data v1.08\x00"
    data += (b"\x00" * 16)
    for i in range(0, 0x1000, 16):
        frame = struct.pack(">cHB", b"R", i, 16)
        radio.pipe.write(frame)
        result = radio.pipe.read(20)
        if frame[1:4] != result[1:4]:
            LOG.debug(util.hexprint(result))
            raise errors.RadioError("Invalid response for address 0x%04x" % i)
        data += result[4:]
        do_status(radio, "from", i)

    return memmap.MemoryMapBytes(data)


def do_upload(radio):
    do_ident(radio)
    data = radio._mmap.get_byte_compatible()[0x0030:]

    for i in range(0, 0x1000, 16):
        frame = struct.pack(">cHB", b"W", i, 16)
        frame += data[i:i + 16]
        radio.pipe.write(frame)
        ack = radio.pipe.read(1)
        if ack != b"\x06":
            # UV-B5/UV-B6 radios with 27 menus do not support service settings
            # and will stop ACKing when the upload reaches 0x0F10
            if i == 0x0F10:
                # must be a radio with 27 menus detected - stop upload
                break
            else:
                LOG.debug("Radio NAK'd block at address 0x%04x" % i)
                raise errors.RadioError(
                    "Radio NAK'd block at address 0x%04x" % i)
        LOG.debug("Radio ACK'd block at address 0x%04x" % i)
        do_status(radio, "to", i)


DUPLEX = ["", "-", "+"]
CHARSET = "0123456789- ABCDEFGHIJKLMNOPQRSTUVWXYZ/_+*"
POWER_LEVELS = [chirp_common.PowerLevel("Low", watts=1),
                chirp_common.PowerLevel("High", watts=5)]


@directory.register
class Th350Radio(BaofengUVB5):
    """TYT TH-350"""
    VENDOR = "TYT"
    MODEL = "TH-350"
    BAUD_RATE = 9600
    NEEDS_COMPAT_SERIAL = False
    SPECIALS = {
        "VFO1": -3,
        "VFO2": -2,
        "VFO3": -1,
    }

    _memsize = 0x1000

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.experimental = \
            ("This TYT TH-350 driver is an alpha version. "
             "Proceed with Caution and backup your data. "
             "Always confirm the correctness of your settings with the "
             "official programmer.")
        return rp

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_settings = True
        rf.has_cross = True
        rf.has_rx_dtcs = True
        rf.valid_tmodes = ["", "Tone", "TSQL", "DTCS", "Cross"]
        rf.valid_cross_modes = ["Tone->Tone", "Tone->DTCS", "DTCS->Tone",
                                "->Tone", "->DTCS", "DTCS->", "DTCS->DTCS"]
        rf.valid_duplexes = DUPLEX + ["split"]
        rf.can_odd_split = True
        rf.valid_skips = ["", "S"]
        rf.valid_characters = CHARSET
        rf.valid_name_length = 5
        rf.valid_bands = [(130000000, 175000000),
                          (220000000, 269000000),
                          (400000000, 520000000)]
        rf.valid_modes = ["FM", "NFM"]
        rf.valid_special_chans = list(self.SPECIALS.keys())
        rf.valid_power_levels = POWER_LEVELS
        rf.valid_tuning_steps = [5.0, 6.25, 10.0, 12.5, 20.0, 25.0]
        rf.has_ctone = True
        rf.has_bank = False
        rf.has_tuning_step = False
        rf.memory_bounds = (1, 128)
        return rf

    def sync_in(self):
        try:
            self._mmap = do_download(self)
        except errors.RadioError:
            raise
        except Exception as e:
            raise errors.RadioError("Failed to communicate with radio: %s" % e)
        self.process_mmap()

    def sync_out(self):
        try:
            do_upload(self)
        except errors.RadioError:
            raise
        except Exception as e:
            raise errors.RadioError("Failed to communicate with radio: %s" % e)

    def process_mmap(self):
        self._memobj = bitwise.parse(mem_format, self._mmap)

    def _decode_tone(self, _mem, which):
        def _get(field):
            return int(getattr(_mem, "%s%s" % (which, field)))

        tonea, toneb = _get('tonea'), _get('toneb')

        if tonea == 0xff:
            mode = val = pol = None
        elif tonea >= 0x80:
            # DTCS
            # D754N -> 0x87 0x54
            # D754I -> 0xc7 0x54
            # Yes. Decimal digits as hex. You're seeing that right.
            # No idea why TYT engineers would do something like that.
            pold = tonea // 16
            if pold not in [0x8, 0xc]:
                LOG.warn("Bug: tone is %04x %04x" % (tonea, toneb))
                mode = val = pol = None
            else:
                mode = 'DTCS'
                val = ((tonea % 16) * 100 +
                       toneb // 16 * 10 +
                       (toneb % 16))
                pol = 'N' if pold == 8 else 'R'
        else:
            # Tone
            # 107.2 -> 0x10 0x72. Seriously.
            mode = 'Tone'
            val = (tonea // 16 * 100 +
                   (tonea % 16) * 10 +
                   toneb // 16 +
                   toneb % 16 / 10)
            pol = None

        return mode, val, pol

    def _encode_tone(self, _mem, which, mode, val, pol):
        def _set(field, value):
            setattr(_mem, "%s%s" % (which, field), value)

        if mode == "Tone":
            tonea = int(
                        val // 100 * 16 +
                        val // 10 % 10
                    )
            toneb = int(
                        floor(val % 10) * 16 +
                        floor(val * 10) % 10
                    )
        elif mode == "DTCS":
            tonea = (0x80 if pol == 'N' else 0xc0) + \
                val // 100
            toneb = (val // 10) % 10 * 16 + \
                val % 10
        else:
            tonea = toneb = 0xff

        _set('tonea', tonea)
        _set('toneb', toneb)

    def _get_memobjs(self, number):
        if isinstance(number, str):
            return (getattr(self._memobj, number.lower()), None)
        elif number < 0:
            for k, v in list(self.SPECIALS.items()):
                if number == v:
                    return (getattr(self._memobj, k.lower()), None)
        else:
            return (self._memobj.channels[number - 1],
                    self._memobj.names[number - 1].name)

    @classmethod
    def match_model(cls, filedata, filename):
        return (filedata.startswith(b"TRI350 Radio Program data") and
                len(filedata) == (cls._memsize + 0x30))
